# Copyright (c) 2016 OVH SAS
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
from unittest import mock

import netaddr
from neutron_lib.services.qos import constants as qos_consts
from oslo_utils import uuidutils

from neutron.agent.linux import bridge_lib
from neutron.agent.linux import ip_lib
from neutron.agent.linux import tc_lib
from neutron.privileged.agent.linux import ip_lib as priv_ip_lib
from neutron.tests.common import net_helpers
from neutron.tests.functional import base as functional_base

TEST_HZ_VALUE = 250
LATENCY = 50
BW_LIMIT = 1024
BURST = 512

BASE_DEV_NAME = "test_tap"


class TcLibTestCase(functional_base.BaseSudoTestCase):

    def create_device(self, name):
        """Create a tuntap with the specified name.

        The device is cleaned up at the end of the test.
        """

        ip = ip_lib.IPWrapper()
        tap_device = ip.add_tuntap(name)
        self.addCleanup(tap_device.link.delete)
        tap_device.link.set_up()

    def test_filters_bandwidth_limit(self):
        device_name = "%s_filters" % BASE_DEV_NAME
        self.create_device(device_name)
        tc = tc_lib.TcCommand(device_name, TEST_HZ_VALUE)

        tc.set_filters_bw_limit(BW_LIMIT, BURST)
        bw_limit, burst = tc.get_filters_bw_limits()
        self.assertEqual(BW_LIMIT, bw_limit)
        self.assertEqual(BURST, burst)

        new_bw_limit = BW_LIMIT + 500
        new_burst = BURST + 50

        tc.update_filters_bw_limit(new_bw_limit, new_burst)
        bw_limit, burst = tc.get_filters_bw_limits()
        self.assertEqual(new_bw_limit, bw_limit)
        self.assertEqual(new_burst, burst)

        tc.delete_filters_bw_limit()
        bw_limit, burst = tc.get_filters_bw_limits()
        self.assertIsNone(bw_limit)
        self.assertIsNone(burst)

    def test_tbf_bandwidth_limit(self):
        device_name = "%s_tbf" % BASE_DEV_NAME
        self.create_device(device_name)
        tc = tc_lib.TcCommand(device_name, TEST_HZ_VALUE)

        tc.set_tbf_bw_limit(BW_LIMIT, BURST, LATENCY)
        bw_limit, burst = tc.get_tbf_bw_limits()
        self.assertEqual(BW_LIMIT, bw_limit)
        self.assertEqual(BURST, burst)

        new_bw_limit = BW_LIMIT + 500
        new_burst = BURST + 50

        tc.set_tbf_bw_limit(new_bw_limit, new_burst, LATENCY)
        bw_limit, burst = tc.get_tbf_bw_limits()
        self.assertEqual(new_bw_limit, bw_limit)
        self.assertEqual(new_burst, burst)

        tc.delete_tbf_bw_limit()
        bw_limit, burst = tc.get_tbf_bw_limits()
        self.assertIsNone(bw_limit)
        self.assertIsNone(burst)


class TcPolicyClassTestCase(functional_base.BaseSudoTestCase):

    def _remove_ns(self, namespace):
        priv_ip_lib.remove_netns(namespace)

    def _create_two_namespaces(self):
        self.ns = ['ns1_' + uuidutils.generate_uuid(),
                   'ns2_' + uuidutils.generate_uuid()]
        self.device = ['int1', 'int2']
        self.mac = []
        self.ip = ['10.100.0.1/24', '10.100.0.2/24']
        list(map(priv_ip_lib.create_netns, self.ns))
        ip_wrapper = ip_lib.IPWrapper(self.ns[0])
        ip_wrapper.add_veth(self.device[0], self.device[1], self.ns[1])
        for i in range(2):
            self.addCleanup(self._remove_ns, self.ns[i])
            ip_device = ip_lib.IPDevice(self.device[i], self.ns[i])
            self.mac.append(ip_device.link.address)
            ip_device.link.set_up()
            ip_device.addr.add(self.ip[i])

    def test_list_tc_policy_class_retrieve_statistics(self):
        statistics = {'bytes', 'packets', 'drop', 'overlimits', 'bps', 'pps',
                      'qlen', 'backlog'}
        self._create_two_namespaces()
        tc_lib.add_tc_qdisc(self.device[0], 'htb', parent='root', handle='1:',
                            namespace=self.ns[0])
        tc_lib.add_tc_policy_class(self.device[0], '1:', '1:10',
                                   max_kbps=1000, burst_kb=900, min_kbps=500,
                                   namespace=self.ns[0])
        tc_lib.add_tc_filter_match_mac(self.device[0], '1:', '1:10',
                                       self.mac[1], namespace=self.ns[0])
        tc_classes = tc_lib.list_tc_policy_class(self.device[0],
                                                 namespace=self.ns[0])
        self.assertEqual(1, len(tc_classes))
        self.assertEqual(statistics, set(tc_classes[0]['stats']))

        bytes = tc_classes[0]['stats']['bytes']
        packets = tc_classes[0]['stats']['packets']
        net_helpers.assert_ping(
            self.ns[1], str(netaddr.IPNetwork(self.ip[0]).ip), count=1)
        tc_classes = tc_lib.list_tc_policy_class(self.device[0],
                                                 namespace=self.ns[0])
        self.assertGreater(tc_classes[0]['stats']['bytes'], bytes)
        self.assertGreater(tc_classes[0]['stats']['packets'], packets)

    def test_add_tc_policy_class_check_min_kbps_values(self):
        def warning_args(rate, min_rate):
            return ('TC HTB class policy rate %(rate)s (bytes/second) is '
                    'lower than the minimum accepted %(min_rate)s '
                    '(bytes/second), for device %(device)s, qdisc '
                    '%(qdisc)s and classid %(classid)s',
                    {'rate': rate, 'min_rate': min_rate, 'classid': '1:10',
                     'device': self.device[0], 'qdisc': '1:'})

        self._create_two_namespaces()
        tc_lib.add_tc_qdisc(self.device[0], 'htb', parent='root', handle='1:',
                            namespace=self.ns[0])

        with mock.patch.object(tc_lib, 'LOG') as mock_log:
            # rate > min_rate: OK
            tc_lib.add_tc_policy_class(self.device[0], '1:', '1:10',
                                       max_kbps=2000, burst_kb=1000,
                                       min_kbps=4, namespace=self.ns[0])
            mock_log.warning.assert_not_called()

            # rate < min_rate: min_rate = 466 with burst = 128000
            tc_lib.add_tc_policy_class(self.device[0], '1:', '1:10',
                                       max_kbps=2000, burst_kb=1000,
                                       min_kbps=3, namespace=self.ns[0])
            mock_log.warning.assert_called_once_with(
                *warning_args(3 * 125, tc_lib._calc_min_rate(1000 * 125)))

            # rate < min_rate: min_rate = 455 with burst = 0.8 ceil = 256000
            mock_log.reset_mock()
            tc_lib.add_tc_policy_class(self.device[0], '1:', '1:10',
                                       max_kbps=2000, burst_kb=None,
                                       min_kbps=5, namespace=self.ns[0])
            min_rate = tc_lib._calc_min_rate(qos_consts.DEFAULT_BURST_RATE *
                                             2000 * 125)
            mock_log.warning.assert_called_once_with(
                *warning_args(5 * 125, min_rate))


class TcFiltersTestCase(functional_base.BaseSudoTestCase):

    def _remove_ns(self, namespace):
        priv_ip_lib.remove_netns(namespace)

    def _create_two_namespaces_connected_using_vxlan(self):
        """Create two namespaces connected with a veth pair and VXLAN

        ---------------------------------    ----------------------------------
        (ns1)                           |    |                            (ns2)
        int1: 10.0.100.1/24 <-----------|----|------------> int2: 10.0.100.2/24
          |                             |    |                              |
          |> int1_vxlan1: 10.0.200.1/24 |    |  int1_vxlan2: 10.0.200.2/24 <|
        ---------------------------------    ----------------------------------
        """
        self.vxlan_id = 100
        self.ns = ['ns1_' + uuidutils.generate_uuid(),
                   'ns2_' + uuidutils.generate_uuid()]
        self.device = ['int1', 'int2']
        self.device_vxlan = ['int_vxlan1', 'int_vxlan2']
        self.mac_vxlan = []
        self.ip = ['10.100.0.1/24', '10.100.0.2/24']
        self.ip_vxlan = ['10.200.0.1/24', '10.200.0.2/24']
        for i, ns in enumerate(self.ns):
            priv_ip_lib.create_netns(ns)
            self.addCleanup(self._remove_ns, ns)
            ip_wrapper = ip_lib.IPWrapper(ns)
            if i == 0:
                ip_wrapper.add_veth(self.device[0], self.device[1], self.ns[1])
            ip_wrapper.add_vxlan(self.device_vxlan[i], self.vxlan_id,
                                 self.device[i])
            ip_device = ip_lib.IPDevice(self.device[i], ns)
            ip_device.link.set_up()
            ip_device.addr.add(self.ip[i])
            ip_device_vxlan = ip_lib.IPDevice(self.device_vxlan[i], ns)
            self.mac_vxlan.append(ip_device_vxlan.link.address)
            ip_device_vxlan.link.set_up()
            ip_device_vxlan.addr.add(self.ip_vxlan[i])

        bridge_lib.FdbInterface.append(
            '00:00:00:00:00:00', self.device_vxlan[0], namespace=self.ns[0],
            dst_ip=str(netaddr.IPNetwork(self.ip[1]).ip))
        bridge_lib.FdbInterface.append(
            '00:00:00:00:00:00', self.device_vxlan[1], namespace=self.ns[1],
            dst_ip=str(netaddr.IPNetwork(self.ip[0]).ip))

    def test_add_tc_filter_vxlan(self):
        # The traffic control is applied on the veth pair device of the first
        # namespace (self.ns[0]). The traffic created from the VXLAN interface
        # when replying to the ping (sent from the other namespace), is
        # encapsulated in a VXLAN frame and goes through the veth pair
        # interface.
        self._create_two_namespaces_connected_using_vxlan()

        tc_lib.add_tc_qdisc(self.device[0], 'htb', parent='root', handle='1:',
                            namespace=self.ns[0])
        classes = tc_lib.list_tc_policy_class(self.device[0],
                                              namespace=self.ns[0])
        self.assertEqual(0, len(classes))

        class_ids = []
        for i in range(1, 10):
            class_id = '1:%s' % i
            class_ids.append(class_id)
            tc_lib.add_tc_policy_class(
                self.device[0], '1:', class_id, namespace=self.ns[0],
                min_kbps=1000, max_kbps=2000, burst_kb=1600)

        # Add a filter for a randomly chosen created class, in the first
        # namespace veth pair device, with the VXLAN MAC address. The traffic
        # from the VXLAN device must go through this chosen class.
        chosen_class_id = random.choice(class_ids)
        tc_lib.add_tc_filter_vxlan(
            self.device[0], '1:', chosen_class_id, self.mac_vxlan[0],
            self.vxlan_id, namespace=self.ns[0])

        tc_classes = tc_lib.list_tc_policy_class(self.device[0],
                                              namespace=self.ns[0])
        for tc_class in (c for c in tc_classes if
                         c['classid'] == chosen_class_id):
            bytes = tc_class['stats']['bytes']
            packets = tc_class['stats']['packets']
            break
        else:
            self.fail('TC class %(class_id)s is not present in the device '
                      '%(device)s' % {'class_id': chosen_class_id,
                                      'device': self.device[0]})

        net_helpers.assert_ping(
            self.ns[1], str(netaddr.IPNetwork(self.ip_vxlan[0]).ip), count=1)
        tc_classes = tc_lib.list_tc_policy_class(self.device[0],
                                                 namespace=self.ns[0])
        for tc_class in tc_classes:
            if tc_class['classid'] == chosen_class_id:
                self.assertGreater(tc_class['stats']['bytes'], bytes)
                self.assertGreater(tc_class['stats']['packets'], packets)
            else:
                self.assertEqual(0, tc_class['stats']['bytes'])
                self.assertEqual(0, tc_class['stats']['packets'])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants
from oslo_utils import uuidutils

from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.unit import testlib_api

load_tests = testlib_api.module_load_tests


class TestPortsApi(base.BaseFullStackTestCase):

    scenarios = [
        ('Sriov Agent', {'l2_agent_type': constants.AGENT_TYPE_NIC_SWITCH})]

    def setUp(self):
        host_descriptions = [
            environment.HostDescription(
                l2_agent_type=self.l2_agent_type)]
        env = environment.Environment(
            environment.EnvironmentDescription(
                agent_down_time=10,
                ml2_extension_drivers=['uplink_status_propagation']),
            host_descriptions)
        super(TestPortsApi, self).setUp(env)

        self.tenant_id = uuidutils.generate_uuid()
        self.network = self.safe_client.create_network(self.tenant_id)
        self.safe_client.create_subnet(
            self.tenant_id, self.network['id'], '20.0.0.0/24')

    def test_create_port_with_propagate_uplink_status(self):
        body = self.safe_client.create_port(
            self.tenant_id, self.network['id'], propagate_uplink_status=False)
        self.assertFalse(body['propagate_uplink_status'])
        body = self.safe_client.client.list_ports(id=body['id'])['ports'][0]
        self.assertFalse(body['propagate_uplink_status'])
        body = self.safe_client.client.show_port(body['id'])['port']
        self.assertFalse(body['propagate_uplink_status'])

    def test_create_port_without_propagate_uplink_status(self):
        body = self.safe_client.create_port(self.tenant_id, self.network['id'])
        self.assertTrue(body['propagate_uplink_status'])
        body = self.safe_client.client.list_ports(id=body['id'])['ports'][0]
        self.assertTrue(body['propagate_uplink_status'])
        body = self.safe_client.client.show_port(body['id'])['port']
        self.assertTrue(body['propagate_uplink_status'])



# Copyright (c) 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_log import log as logging
from oslo_service import service
from ovsdbapp.backend.ovs_idl import event as row_event

from neutron.agent.ovn.agent import ovsdb
from neutron.agent.ovn.extensions import extension_manager as ext_mgr
from neutron.common.ovn import constants as ovn_const


LOG = logging.getLogger(__name__)
OVN_MONITOR_UUID_NAMESPACE = uuid.UUID('fd7e0970-7164-11ed-80f0-00000003158a')


class SbGlobalUpdateEvent(row_event.RowEvent):
    """Row update event on SB_Global table.

    This event will trigger the OVN Neutron Agent update of the
    'neutron:ovn-neutron-agent-sb-cfg' key in 'SB_Global', that is used to
    determine the agent status.
    """

    def __init__(self, ovn_agent):
        self.ovn_agent = ovn_agent
        table = 'SB_Global'
        events = (self.ROW_UPDATE, )
        super().__init__(events, table, None)
        self.event_name = self.__class__.__name__

    def run(self, event, row, old):
        ext_ids = {ovn_const.OVN_AGENT_NEUTRON_SB_CFG_KEY: str(row.nb_cfg)}
        self.ovn_agent.sb_idl.db_set('Chassis_Private', self.ovn_agent.chassis,
                                     ('external_ids', ext_ids)).execute()


class OVNNeutronAgent(service.Service):

    def __init__(self, conf):
        super().__init__()
        self._conf = conf
        self.chassis = None
        self.chassis_id = None
        self.ovn_bridge = None
        self.ext_manager_api = ext_mgr.OVNAgentExtensionAPI()
        self.ext_manager = ext_mgr.OVNAgentExtensionManager(self._conf)
        self.ext_manager.initialize(None, 'ovn', self)

    def __getitem__(self, name):
        """Return the named extension objet from ``self.ext_manager``"""
        return self.ext_manager[name].obj

    @property
    def ovs_idl(self):
        return self.ext_manager_api.ovs_idl

    @property
    def nb_idl(self):
        return self.ext_manager_api.nb_idl

    @property
    def nb_post_fork_event(self):
        return self.ext_manager_api.nb_post_fork_event

    @property
    def sb_idl(self):
        return self.ext_manager_api.sb_idl

    @property
    def sb_post_fork_event(self):
        return self.ext_manager_api.sb_post_fork_event

    def _load_config(self, ovs_idl):
        self.chassis = ovsdb.get_own_chassis_name(ovs_idl)
        try:
            self.chassis_id = uuid.UUID(self.chassis)
        except ValueError:
            # OVS system-id could be a non UUID formatted string.
            self.chassis_id = uuid.uuid5(OVN_MONITOR_UUID_NAMESPACE,
                                         self.chassis)
        self.ovn_bridge = ovsdb.get_ovn_bridge(ovs_idl)
        LOG.info("Loaded chassis name %s (UUID: %s) and ovn bridge %s.",
                 self.chassis, self.chassis_id, self.ovn_bridge)

    def _load_ovs_idl(self):
        events = []
        for extension in self.ext_manager:
            events += extension.obj.ovs_idl_events
        events = [e(self) for e in set(events)]
        return ovsdb.MonitorAgentOvsIdl(set(events)).start()

    def _load_nb_idl(self):
        events = []
        tables = []
        for extension in self.ext_manager:
            events += extension.obj.nb_idl_events
            tables += extension.obj.nb_idl_tables

        if not (tables or events):
            # If there is no need to retrieve any table nor attend to any
            # event, the IDL object is not created to save a DB connection.
            return None

        events = [e(self) for e in set(events)]
        tables = set(tables)
        return ovsdb.MonitorAgentOvnNbIdl(tables, events).start()

    def _load_sb_idl(self):
        events = [SbGlobalUpdateEvent]
        tables = ['SB_Global', 'Chassis_Private']
        for extension in self.ext_manager:
            events += extension.obj.sb_idl_events
            tables += extension.obj.sb_idl_tables

        events = [e(self) for e in set(events)]
        tables = set(tables)
        return ovsdb.MonitorAgentOvnSbIdl(tables, events,
                                          chassis=self.chassis).start()

    def start(self):
        self.ext_manager_api.ovs_idl = self._load_ovs_idl()
        # Before executing "_load_config", it is needed to create the OVS IDL.
        self._load_config(self.ext_manager_api.ovs_idl)
        # Before executing "_load_sb_idl", is is needed to execute
        # "_load_config" to populate self.chassis.
        self.ext_manager_api.sb_idl = self._load_sb_idl()
        self.ext_manager_api.nb_idl = self._load_nb_idl()
        LOG.info('Starting OVN Neutron Agent')

    def stop(self, graceful=True):
        LOG.info('Stopping OVN Neutron Agent')
        super().stop(graceful)



# Copyright 2016 Hewlett Packard Enterprise Development, LP
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from keystoneauth1 import loading as ks_loading
import netaddr
from neutron_lib.api.definitions import ip_allocation as ipalloc_apidef
from neutron_lib.api.definitions import l2_adjacency as l2adj_apidef
from neutron_lib.api.definitions import network as net_def
from neutron_lib.api.definitions import port as port_def
from neutron_lib.api.definitions import segment as seg_apidef
from neutron_lib.api.definitions import segments_peer_subnet_host_routes
from neutron_lib.api.definitions import standard_attr_segment
from neutron_lib.api.definitions import subnet as subnet_def
from neutron_lib.api.definitions import subnet_segmentid_writable
from neutron_lib.api import validators
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.db import resource_extend
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import placement as placement_exc
from neutron_lib.placement import client as placement_client
from neutron_lib.plugins import directory
from novaclient import client as nova_client
from novaclient import exceptions as nova_exc
from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils

from neutron._i18n import _
from neutron.common import ipv6_utils
from neutron.extensions import segment
from neutron.notifiers import batch_notifier
from neutron.objects import network as net_obj
from neutron.objects import ports as ports_obj
from neutron.objects import subnet as subnet_obj
from neutron.services.segments import db
from neutron.services.segments import exceptions


LOG = log.getLogger(__name__)

NOVA_API_VERSION = '2.41'
IPV4_RESOURCE_CLASS = 'IPV4_ADDRESS'
SEGMENT_NAME_STUB = 'Neutron segment id %s'
MAX_INVENTORY_UPDATE_RETRIES = 10


@resource_extend.has_resource_extenders
@registry.has_registry_receivers
class Plugin(db.SegmentDbMixin, segment.SegmentPluginBase):

    _instance = None

    supported_extension_aliases = [seg_apidef.ALIAS,
                                   ipalloc_apidef.ALIAS,
                                   l2adj_apidef.ALIAS,
                                   standard_attr_segment.ALIAS,
                                   subnet_segmentid_writable.ALIAS,
                                   segments_peer_subnet_host_routes.ALIAS]

    __native_pagination_support = True
    __native_sorting_support = True
    __filter_validation_support = True

    def __init__(self):
        self.nova_updater = NovaSegmentNotifier()
        self.segment_host_routes = SegmentHostRoutes()

    @staticmethod
    @resource_extend.extends([net_def.COLLECTION_NAME])
    def _extend_network_dict_binding(network_res, network_db):
        if not directory.get_plugin('segments'):
            return

        # TODO(carl_baldwin) Make this work with service subnets when
        #                    it's a thing.
        is_adjacent = (not network_db.subnets or
                       not network_db.subnets[0].segment_id)
        network_res[l2adj_apidef.L2_ADJACENCY] = is_adjacent

    @staticmethod
    @resource_extend.extends([subnet_def.COLLECTION_NAME])
    def _extend_subnet_dict_binding(subnet_res, subnet_db):
        subnet_res['segment_id'] = subnet_db.get('segment_id')

    @staticmethod
    @resource_extend.extends([port_def.COLLECTION_NAME])
    def _extend_port_dict_binding(port_res, port_db):
        if not directory.get_plugin('segments'):
            return

        value = ipalloc_apidef.IP_ALLOCATION_IMMEDIATE
        if port_db.get('ip_allocation'):
            value = port_db.get('ip_allocation')
        port_res[ipalloc_apidef.IP_ALLOCATION] = value

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @registry.receives(resources.SEGMENT, [events.BEFORE_DELETE])
    def _prevent_segment_delete_with_subnet_associated(
            self, resource, event, trigger, payload=None):
        """Raise exception if there are any subnets associated with segment."""
        if payload.metadata.get(db.FOR_NET_DELETE):
            # don't check if this is a part of a network delete operation
            return
        segment_id = payload.resource_id
        subnets = subnet_obj.Subnet.get_objects(payload.context,
                                                segment_id=segment_id)
        subnet_ids = [s.id for s in subnets]

        if subnet_ids:
            reason = _("The segment is still associated with subnet(s) "
                       "%s") % ", ".join(subnet_ids)
            raise exceptions.SegmentInUse(segment_id=segment_id,
                                          reason=reason)

    @registry.receives(
        resources.SUBNET, [events.PRECOMMIT_DELETE_ASSOCIATIONS])
    def _validate_auto_address_subnet_delete(self, resource, event, trigger,
                                             payload):
        context = payload.context
        subnet = subnet_obj.Subnet.get_object(context, id=payload.resource_id)
        is_auto_addr_subnet = ipv6_utils.is_auto_address_subnet(subnet)
        if not is_auto_addr_subnet or subnet.segment_id is None:
            return

        ports = ports_obj.Port.get_ports_allocated_by_subnet_id(context,
                                                                subnet.id)
        for port in ports:
            fixed_ips = [f for f in port.fixed_ips if f.subnet_id != subnet.id]
            if len(fixed_ips) != 0:
                continue

            LOG.info("Found port %(port_id)s, with IP auto-allocation "
                     "only on subnet %(subnet)s which is associated with "
                     "segment %(segment_id)s, cannot delete",
                     {'port_id': port.id,
                      'subnet': subnet.id,
                      'segment_id': subnet.segment_id})
            raise n_exc.SubnetInUse(subnet_id=subnet.id)


class Event(object):

    def __init__(self, method, segment_ids, total=None, reserved=None,
                 segment_host_mappings=None, host=None):
        self.method = method
        if isinstance(segment_ids, set):
            self.segment_ids = segment_ids
        else:
            self.segment_id = segment_ids
        self.total = total
        self.reserved = reserved
        self.segment_host_mappings = segment_host_mappings
        self.host = host


@registry.has_registry_receivers
class NovaSegmentNotifier(object):

    def __init__(self):
        self.p_client, self.n_client = self._get_clients()
        self.batch_notifier = batch_notifier.BatchNotifier(
            cfg.CONF.send_events_interval, self._send_notifications)

    def _get_clients(self):
        p_client = placement_client.PlacementAPIClient(
            cfg.CONF, openstack_api_version='placement 1.1')

        n_auth = ks_loading.load_auth_from_conf_options(cfg.CONF, 'nova')
        n_session = ks_loading.load_session_from_conf_options(
            cfg.CONF,
            'nova',
            auth=n_auth)
        extensions = [
            ext for ext in nova_client.discover_extensions(NOVA_API_VERSION)
            if ext.name == "server_external_events"]
        n_client = nova_client.Client(
            NOVA_API_VERSION,
            session=n_session,
            region_name=cfg.CONF.nova.region_name,
            endpoint_type=cfg.CONF.nova.endpoint_type,
            extensions=extensions)

        return p_client, n_client

    def _send_notifications(self, batched_events):
        for event in batched_events:
            try:
                event.method(event)
            except placement_exc.PlacementEndpointNotFound:
                LOG.debug('Placement API was not found when trying to '
                          'update routed networks IPv4 inventories')
                return

    def _notify_subnet(self, context, subnet, segment_id):
        total, reserved = self._calculate_inventory_total_and_reserved(subnet)
        if total:
            segment_host_mappings = net_obj.SegmentHostMapping.get_objects(
                context, segment_id=segment_id)
            self.batch_notifier.queue_event(Event(
                self._create_or_update_nova_inventory, segment_id, total=total,
                reserved=reserved,
                segment_host_mappings=segment_host_mappings))

    @registry.receives(resources.SUBNET, [events.AFTER_CREATE])
    def _notify_subnet_created(self, resource, event, trigger, payload):
        context = payload.context
        subnet = payload.latest_state
        segment_id = subnet.get('segment_id')
        if not segment_id or subnet['ip_version'] != constants.IP_VERSION_4:
            return
        self._notify_subnet(context, subnet, segment_id)

    def _create_or_update_nova_inventory(self, event):
        try:
            self._update_nova_inventory(event)
        except placement_exc.PlacementResourceProviderNotFound:
            self._create_nova_inventory(event.segment_id, event.total,
                                        event.reserved,
                                        event.segment_host_mappings)

    def _update_nova_inventory(self, event):
        for count in range(MAX_INVENTORY_UPDATE_RETRIES):
            ipv4_inventory = self.p_client.get_inventory(event.segment_id,
                                                         IPV4_RESOURCE_CLASS)
            if event.total:
                ipv4_inventory['total'] += event.total
            if event.reserved:
                ipv4_inventory['reserved'] += event.reserved
            try:
                self.p_client.update_resource_provider_inventory(
                    event.segment_id, ipv4_inventory, IPV4_RESOURCE_CLASS)
                return
            except placement_exc.PlacementResourceProviderGenerationConflict:
                LOG.debug('Re-trying to update Nova IPv4 inventory for '
                          'routed network segment: %s', event.segment_id)
        LOG.error('Failed to update Nova IPv4 inventory for routed '
                  'network segment: %s', event.segment_id)

    def _get_nova_aggregate_uuid(self, aggregate):
        try:
            return aggregate.uuid
        except AttributeError:
            with excutils.save_and_reraise_exception():
                LOG.exception("uuid was not returned as part of the aggregate "
                              "object which indicates that the Nova API "
                              "backend does not support microversions. Ensure "
                              "that the compute endpoint in the service "
                              "catalog points to the v2.1 API.")

    def _create_nova_inventory(self, segment_id, total, reserved,
                               segment_host_mappings):
        name = SEGMENT_NAME_STUB % segment_id
        resource_provider = {'name': name, 'uuid': segment_id}
        self.p_client.create_resource_provider(resource_provider)
        aggregate = self.n_client.aggregates.create(name, None)
        aggregate_uuid = self._get_nova_aggregate_uuid(aggregate)
        self.p_client.associate_aggregates(segment_id, [aggregate_uuid])
        for mapping in segment_host_mappings:
            self.n_client.aggregates.add_host(aggregate.id, mapping.host)
        ipv4_inventory = {
            IPV4_RESOURCE_CLASS: {
                'total': total, 'reserved': reserved, 'min_unit': 1,
                'max_unit': 1, 'step_size': 1, 'allocation_ratio': 1.0,
            }
        }
        self.p_client.update_resource_provider_inventories(
            segment_id, ipv4_inventory)

    def _calculate_inventory_total_and_reserved(self, subnet):
        total = 0
        reserved = 0
        allocation_pools = subnet.get('allocation_pools') or []
        for pool in allocation_pools:
            total += int(netaddr.IPAddress(pool['end']) -
                         netaddr.IPAddress(pool['start'])) + 1
        if total:
            if subnet.get('gateway_ip'):
                total += 1
                reserved += 1
            if subnet.get('enable_dhcp'):
                reserved += 1
        return total, reserved

    @registry.receives(resources.SUBNET, [events.AFTER_UPDATE])
    def _notify_subnet_updated(self, resource, event, trigger, payload):
        context = payload.context
        original_subnet = payload.states[0]
        subnet = payload.latest_state
        segment_id = subnet.get('segment_id')
        original_segment_id = original_subnet.get('segment_id')
        if not segment_id or subnet['ip_version'] != constants.IP_VERSION_4:
            return
        if original_segment_id != segment_id:
            # Migration to routed network, treat as create
            self._notify_subnet(context, subnet, segment_id)
            return
        filters = {'segment_id': [segment_id],
                   'ip_version': [constants.IP_VERSION_4]}
        if not subnet['allocation_pools']:
            plugin = directory.get_plugin()
            alloc_pools = [s['allocation_pools'] for s in
                           plugin.get_subnets(context, filters=filters)]
            if not any(alloc_pools):
                self.batch_notifier.queue_event(Event(
                    self._delete_nova_inventory, segment_id))
                return
        original_total, original_reserved = (
            self._calculate_inventory_total_and_reserved(original_subnet))
        updated_total, updated_reserved = (
            self._calculate_inventory_total_and_reserved(subnet))
        total = updated_total - original_total
        reserved = updated_reserved - original_reserved
        if total or reserved:
            segment_host_mappings = None
            if not original_subnet['allocation_pools']:
                segment_host_mappings = net_obj.SegmentHostMapping.get_objects(
                    context, segment_id=segment_id)
            self.batch_notifier.queue_event(Event(
                self._create_or_update_nova_inventory, segment_id, total=total,
                reserved=reserved,
                segment_host_mappings=segment_host_mappings))

    @registry.receives(resources.SUBNET, [events.AFTER_DELETE])
    def _notify_subnet_deleted(self, resource, event, trigger, payload):
        context = payload.context
        subnet = payload.latest_state
        if payload.metadata.get(db.FOR_NET_DELETE):
            return  # skip segment RP update if it is going to be deleted

        segment_id = subnet.get('segment_id')
        if not segment_id or subnet['ip_version'] != constants.IP_VERSION_4:
            return
        total, reserved = self._calculate_inventory_total_and_reserved(subnet)
        if total:
            filters = {'segment_id': [segment_id], 'ip_version': [4]}
            plugin = directory.get_plugin()
            if plugin.get_subnets_count(context, filters=filters) > 0:
                self.batch_notifier.queue_event(Event(
                    self._update_nova_inventory, segment_id, total=-total,
                    reserved=-reserved))
            else:
                self.batch_notifier.queue_event(Event(
                    self._delete_nova_inventory, segment_id))

    def _get_aggregate_id(self, segment_id):
        try:
            aggregate_uuid = self.p_client.list_aggregates(
                segment_id)['aggregates'][0]
        except placement_exc.PlacementAggregateNotFound:
            LOG.info('Segment %s resource provider aggregate not found',
                     segment_id)
            return

        for aggregate in self.n_client.aggregates.list():
            nc_aggregate_uuid = self._get_nova_aggregate_uuid(aggregate)
            if nc_aggregate_uuid == aggregate_uuid:
                return aggregate.id

    def _delete_nova_inventory(self, event):
        aggregate_id = self._get_aggregate_id(event.segment_id)
        if aggregate_id:
            aggregate = self.n_client.aggregates.get_details(aggregate_id)
            for host in aggregate.hosts:
                self.n_client.aggregates.remove_host(aggregate_id, host)
            self.n_client.aggregates.delete(aggregate_id)

        try:
            self.p_client.delete_resource_provider(event.segment_id)
        except placement_exc.PlacementClientError as exc:
            LOG.info('Segment %s resource provider not found; error: %s',
                     event.segment_id, str(exc))

    @registry.receives(resources.SEGMENT_HOST_MAPPING, [events.AFTER_CREATE])
    def _notify_host_addition_to_aggregate(self, resource, event, trigger,
                                           payload=None):
        subnets = subnet_obj.Subnet.get_objects(
            payload.context,
            segment_id=payload.metadata.get('current_segment_ids'))
        segment_ids = {s.segment_id for s in subnets}
        self.batch_notifier.queue_event(
            Event(self._add_host_to_aggregate,
                  segment_ids, host=payload.metadata.get('host')))

    def _add_host_to_aggregate(self, event):
        for segment_id in event.segment_ids:
            aggregate_id = self._get_aggregate_id(segment_id)
            if not aggregate_id:
                LOG.info('When adding host %(host)s, aggregate not found '
                         'for routed network segment %(segment_id)s',
                         {'host': event.host, 'segment_id': segment_id})
                continue

            try:
                self.n_client.aggregates.add_host(aggregate_id, event.host)
            except nova_exc.Conflict:
                LOG.info('Host %(host)s already exists in aggregate for '
                         'routed network segment %(segment_id)s',
                         {'host': event.host, 'segment_id': segment_id})

    @registry.receives(resources.PORT, [events.AFTER_CREATE,
                                        events.AFTER_DELETE])
    def _notify_port_created_or_deleted(self, resource, event, trigger,
                                        payload):
        context = payload.context
        port = payload.latest_state
        if not self._does_port_require_nova_inventory_update(port):
            return
        ipv4_subnets_number, segment_id = (
            self._get_ipv4_subnets_number_and_segment_id(port, context))
        if segment_id:
            if event == events.AFTER_DELETE:
                ipv4_subnets_number = -ipv4_subnets_number
            self.batch_notifier.queue_event(
                Event(self._update_nova_inventory,
                      segment_id, reserved=ipv4_subnets_number))

    @registry.receives(resources.PORT, [events.AFTER_UPDATE])
    def _notify_port_updated(self, resource, event, trigger, payload):
        context = payload.context
        port = payload.latest_state
        original_port = payload.states[0]
        does_original_port_require_nova_inventory_update = (
            self._does_port_require_nova_inventory_update(original_port))
        does_port_require_nova_inventory_update = (
            self._does_port_require_nova_inventory_update(port))
        if not (does_original_port_require_nova_inventory_update or
                does_port_require_nova_inventory_update):
            return
        original_port_ipv4_subnets_number, segment_id = (
            self._get_ipv4_subnets_number_and_segment_id(original_port,
                                                         context))
        if not segment_id:
            return
        port_ipv4_subnets_number = len(self._get_ipv4_subnet_ids(port))
        if not does_original_port_require_nova_inventory_update:
            original_port_ipv4_subnets_number = 0
        if not does_port_require_nova_inventory_update:
            port_ipv4_subnets_number = 0
        update = port_ipv4_subnets_number - original_port_ipv4_subnets_number
        if update:
            self.batch_notifier.queue_event(Event(self._update_nova_inventory,
                                                  segment_id, reserved=update))

    def _get_ipv4_subnets_number_and_segment_id(self, port, context):
        ipv4_subnet_ids = self._get_ipv4_subnet_ids(port)
        if not ipv4_subnet_ids:
            return 0, None
        subnet = subnet_obj.Subnet.get_object(context, id=ipv4_subnet_ids[0])
        if subnet and subnet.segment_id:
            return len(ipv4_subnet_ids), subnet.segment_id

        return 0, None

    def _does_port_require_nova_inventory_update(self, port):
        device_owner = port.get('device_owner')
        if (device_owner.startswith(constants.DEVICE_OWNER_COMPUTE_PREFIX) or
                device_owner == constants.DEVICE_OWNER_DHCP):
            return False
        return True

    def _get_ipv4_subnet_ids(self, port):
        ipv4_subnet_ids = []
        for ip in port.get('fixed_ips', []):
            if netaddr.IPAddress(
                    ip['ip_address']).version == constants.IP_VERSION_4:
                ipv4_subnet_ids.append(ip['subnet_id'])
        return ipv4_subnet_ids

    @registry.receives(resources.SEGMENT, [events.AFTER_DELETE])
    def _notify_segment_deleted(
            self, resource, event, trigger, payload=None):
        if payload:
            self.batch_notifier.queue_event(Event(
                self._delete_nova_inventory, payload.resource_id))


@registry.has_registry_receivers
class SegmentHostRoutes(object):

    def _get_subnets(self, context, network_id):
        return subnet_obj.Subnet.get_objects(context, network_id=network_id)

    def _count_subnets(self, context, network_id):
        return subnet_obj.Subnet.count(context, network_id=network_id)

    def _calculate_routed_network_host_routes(self, context, ip_version,
                                              network_id=None, subnet_id=None,
                                              segment_id=None,
                                              host_routes=None,
                                              gateway_ip=None,
                                              old_gateway_ip=None,
                                              deleted_cidr=None):
        """Calculate host routes for routed network.

        This method is used to calculate the host routes for routed networks
        both when handling the user create or update request and when making
        updates to subnets on the network in response to events: AFTER_CREATE
        and AFTER_DELETE.

        :param ip_version: IP version (4/6).
        :param network_id: Network ID.
        :param subnet_id: UUID of the subnet.
        :param segment_id: Segement ID associated with the subnet.
        :param host_routes: Current host_routes of the subnet.
        :param gateway_ip: The subnets gateway IP address.
        :param old_gateway_ip: The old gateway IP address of the subnet when it
                               is changed on update.
        :param deleted_cidr: The cidr of a deleted subnet.
        :returns Host routes with routes for the other subnet's on the routed
                 network appended unless a route to the destination already
                 exists.
        """
        if host_routes is None:
            host_routes = []
        dest_ip_nets = [netaddr.IPNetwork(route['destination']) for
                        route in host_routes]

        # Drop routes to the deleted cidr, when the subnet was deleted.
        if deleted_cidr:
            delete_route = {'destination': deleted_cidr, 'nexthop': gateway_ip}
            if delete_route in host_routes:
                host_routes.remove(delete_route)

        for subnet in self._get_subnets(context, network_id):
            if (subnet.id == subnet_id or subnet.segment_id == segment_id or
                    subnet.ip_version != ip_version):
                continue
            subnet_ip_net = netaddr.IPNetwork(subnet.cidr)
            if old_gateway_ip:
                old_route = {'destination': str(subnet.cidr),
                             'nexthop': old_gateway_ip}
                if old_route in host_routes:
                    host_routes.remove(old_route)
                    dest_ip_nets.remove(subnet_ip_net)
            if gateway_ip:
                # Use netaddr here in case the user provided a summary route
                # (supernet route). I.e subnet.cidr = 10.0.1.0/24 and
                # the user provided a host route for 10.0.0.0/16. We don't
                # need to append a route in this case.
                if not any(subnet_ip_net in ip_net for ip_net in dest_ip_nets):
                    host_routes.append({'destination': subnet.cidr,
                                        'nexthop': gateway_ip})

        return host_routes

    def _host_routes_need_update(self, host_routes, calc_host_routes):
        """Compare host routes and calculated host routes

        :param host_routes: Current host routes
        :param calc_host_routes: Host routes + calculated host routes for
                                 routed network
        :returns True if host_routes and calc_host_routes are not equal
        """
        return ((set((route['destination'],
                      route['nexthop']) for route in host_routes) !=
                 set((route['destination'],
                      route['nexthop']) for route in calc_host_routes)))

    def _update_routed_network_host_routes(self, context, network_id,
                                           deleted_cidr=None):
        """Update host routes on subnets on a routed network after event

        Host routes on the subnets on a routed network may need updates after
        any CREATE or DELETE event.

        :param network_id: Network ID
        :param deleted_cidr: The cidr of a deleted subnet.
        """
        for subnet in self._get_subnets(context, network_id):
            host_routes = [{'destination': str(route.destination),
                            'nexthop': route.nexthop}
                           for route in subnet.host_routes]
            calc_host_routes = self._calculate_routed_network_host_routes(
                context=context,
                ip_version=subnet.ip_version,
                network_id=subnet.network_id,
                subnet_id=subnet.id,
                segment_id=subnet.segment_id,
                host_routes=copy.deepcopy(host_routes),
                gateway_ip=subnet.gateway_ip,
                deleted_cidr=deleted_cidr)
            if self._host_routes_need_update(host_routes, calc_host_routes):
                LOG.debug(
                    "Updating host routes for subnet %s on routed network %s",
                    subnet.id, subnet.network_id)
                plugin = directory.get_plugin()
                plugin.update_subnet(context, subnet.id,
                                     {'subnet': {
                                         'host_routes': calc_host_routes}})

    @registry.receives(resources.SUBNET, [events.BEFORE_CREATE])
    def host_routes_before_create(self, resource, event, trigger,
                                  payload):
        context = payload.context
        subnet = payload.latest_state
        segment_id = subnet.get('segment_id')
        gateway_ip = subnet.get('gateway_ip')
        if validators.is_attr_set(subnet.get('host_routes')):
            host_routes = subnet.get('host_routes')
        else:
            host_routes = []
        if segment_id is not None and validators.is_attr_set(gateway_ip):
            calc_host_routes = self._calculate_routed_network_host_routes(
                context=context,
                ip_version=netaddr.IPNetwork(subnet['cidr']).version,
                network_id=subnet['network_id'],
                segment_id=segment_id,
                host_routes=copy.deepcopy(host_routes),
                gateway_ip=gateway_ip)
            if (not host_routes or
                    self._host_routes_need_update(host_routes,
                                                  calc_host_routes)):
                subnet['host_routes'] = calc_host_routes

    @registry.receives(resources.SUBNET, [events.BEFORE_UPDATE])
    def host_routes_before_update(self, resource, event, trigger,
                                  payload):
        context = payload.context
        original_subnet = payload.states[0]
        subnet = payload.latest_state
        orig_segment_id = original_subnet.get('segment_id')
        segment_id = subnet.get('segment_id', orig_segment_id)
        orig_gateway_ip = original_subnet.get('gateway_ip')
        gateway_ip = subnet.get('gateway_ip', orig_gateway_ip)
        orig_host_routes = original_subnet.get('host_routes')
        host_routes = subnet.get('host_routes', orig_host_routes)
        if (segment_id and (host_routes != orig_host_routes or
                            gateway_ip != orig_gateway_ip)):
            calc_host_routes = self._calculate_routed_network_host_routes(
                context=context,
                ip_version=netaddr.IPNetwork(original_subnet['cidr']).version,
                network_id=original_subnet['network_id'],
                segment_id=segment_id,
                host_routes=copy.deepcopy(host_routes),
                gateway_ip=gateway_ip,
                old_gateway_ip=orig_gateway_ip if (
                    gateway_ip != orig_gateway_ip) else None)
            if self._host_routes_need_update(host_routes, calc_host_routes):
                subnet['host_routes'] = calc_host_routes

    @registry.receives(resources.SUBNET, [events.AFTER_CREATE])
    def host_routes_after_create(self, resource, event, trigger,
                                 payload):
        context = payload.context
        subnet = payload.latest_state
        # If there are other subnets on the network and subnet has segment_id
        # ensure host routes for all subnets are updated.

        if (subnet.get('segment_id') and
                self._count_subnets(context, subnet['network_id']) > 1):
            self._update_routed_network_host_routes(context,
                                                    subnet['network_id'])

    @registry.receives(resources.SUBNET, [events.AFTER_DELETE])
    def host_routes_after_delete(self, resource, event, trigger,
                                 payload):
        # If this is a routed network, remove any routes to this subnet on
        # this networks remaining subnets.
        context = payload.context
        subnet = payload.latest_state
        if payload.metadata.get(db.FOR_NET_DELETE):
            return  # skip subnet update if the network is going to be deleted

        if subnet.get('segment_id'):
            self._update_routed_network_host_routes(
                context, subnet['network_id'], deleted_cidr=subnet['cidr'])



# Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import ovs_constants as agent_consts
from neutron_lib.services.trunk import constants as trunk_consts
from oslo_config import cfg
from oslo_log import log as logging

from neutron.objects import trunk as trunk_obj
from neutron.services.trunk.drivers import base
from neutron.services.trunk.drivers.openvswitch import utils
from neutron.services.trunk import exceptions as trunk_exc

LOG = logging.getLogger(__name__)

NAME = 'openvswitch'

SUPPORTED_INTERFACES = (
    portbindings.VIF_TYPE_OVS,
    portbindings.VIF_TYPE_VHOST_USER,
)

SUPPORTED_SEGMENTATION_TYPES = (
    trunk_consts.SEGMENTATION_TYPE_VLAN,
)

DRIVER = None


class OVSDriver(base.DriverBase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._core_plugin = None

    @property
    def core_plugin(self):
        if not self._core_plugin:
            self._core_plugin = directory.get_plugin()
        return self._core_plugin

    @property
    def is_loaded(self):
        try:
            return NAME in cfg.CONF.ml2.mechanism_drivers
        except cfg.NoSuchOptError:
            return False

    @classmethod
    def create(cls):
        return OVSDriver(NAME,
                         SUPPORTED_INTERFACES,
                         SUPPORTED_SEGMENTATION_TYPES,
                         constants.AGENT_TYPE_OVS)

    @staticmethod
    def _get_trunk(context, trunk_id):
        """Return the trunk object or raise if not found."""
        obj = trunk_obj.Trunk.get_object(context, id=trunk_id)
        if obj is None:
            raise trunk_exc.TrunkNotFound(trunk_id=trunk_id)

        return obj

    def _update_subport_binding(self, context, trunk_id):
        """Update the subport binding host"""
        trunk_obj = self._get_trunk(context, trunk_id)
        trunk_port = self.core_plugin.get_port(context, trunk_obj.port_id)
        trunk_host = trunk_port.get(portbindings.HOST_ID)
        for subport in trunk_obj.sub_ports:
            port = self.core_plugin.update_port(
                context, subport.port_id,
                {'port': {portbindings.HOST_ID: trunk_host,
                          'device_owner': trunk_consts.TRUNK_SUBPORT_OWNER}})
            vif_type = port.get(portbindings.VIF_TYPE)
            if vif_type == portbindings.VIF_TYPE_BINDING_FAILED:
                raise trunk_exc.SubPortBindingError(
                    port_id=subport.port_id, trunk_id=trunk_obj.id)

    @registry.receives(resources.PORT, [events.AFTER_UPDATE])
    def _subport_binding(self, resource, event, trigger, payload=None):
        """Bind the subports to the parent port host

        This method listen to the port after update events. If the parent port
        is updated and transitions from inactive to active, this method
        retrieves the port host ID and binds the subports to this host.

        :param resource: neutron_lib.callbacks.resources.PORT
        :param event: neutron_lib.callbacks.events.AFTER_UPDATE
        :param trigger: the specific driver plugin

        """
        updated_port = payload.latest_state
        trunk_details = updated_port.get('trunk_details')
        vif_details = updated_port.get(portbindings.VIF_DETAILS)
        driver = vif_details.get(portbindings.VIF_DETAILS_BOUND_DRIVERS,
                                 {}).get('0')
        # If no trunk_details, the port is not the parent of a trunk.
        # If this port is not bound to ML2/OVS, we skip this method.
        if not trunk_details or not driver == NAME:
            return

        context = payload.context
        orig_status = payload.states[0].get('status')
        new_status = updated_port.get('status')
        trunk_id = trunk_details['trunk_id']
        if (new_status == constants.PORT_STATUS_ACTIVE and
                new_status != orig_status):
            self._update_subport_binding(context, trunk_id)


def register():
    """Register the driver."""
    global DRIVER
    DRIVER = OVSDriver.create()
    # To set the bridge_name in a parent port's vif_details.
    registry.subscribe(vif_details_bridge_name_handler,
                       agent_consts.OVS_BRIDGE_NAME,
                       events.BEFORE_READ)
    LOG.debug('Open vSwitch trunk driver registered')


def vif_details_bridge_name_handler(resource, event, set_br_name,
                                    payload=None):
    """If port is a trunk port, generate a bridge_name for its vif_details."""
    port = payload.metadata['port']
    if 'trunk_details' in port:
        set_br_name(utils.gen_trunk_br_name(port['trunk_details']['trunk_id']))



# Copyright 2020 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib import constants
from neutron_lib.services.qos import constants as qos_constants
from ovsdbapp.backend.ovs_idl import idlutils

from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.common import utils
from neutron.db import l3_db
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb.extensions \
    import qos as qos_extension
from neutron.tests.functional import base


QOS_RULE_BW_1 = {'max_kbps': 200, 'max_burst_kbps': 100}
QOS_RULE_BW_2 = {'max_kbps': 300}
QOS_RULE_DSCP_1 = {'dscp_mark': 16}
QOS_RULE_DSCP_2 = {'dscp_mark': 20}
QOS_RULE_MINBW_1 = {'min_kbps': 500}

QOS_RULES_0 = {
    constants.EGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1,
        qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_1,
        qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH: QOS_RULE_MINBW_1},
    constants.INGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2}
}

QOS_RULES_1 = {
    constants.EGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1,
        qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_1},
    constants.INGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2}
}

QOS_RULES_2 = {
    constants.EGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_2,
        qos_constants.RULE_TYPE_DSCP_MARKING: QOS_RULE_DSCP_2}
}

QOS_RULES_3 = {
    constants.INGRESS_DIRECTION: {
        qos_constants.RULE_TYPE_BANDWIDTH_LIMIT: QOS_RULE_BW_1}
}


class TestOVNClientQosExtensionBase(base.TestOVNFunctionalBase):

    def _check_rules(self, rules, port_id, network_id, fip_id=None,
                     ip_address=None, check_min_rate=True,
                     expected_ext_ids=None):
        egress_ovn_rule = self.qos_driver._ovn_qos_rule(
            constants.EGRESS_DIRECTION, rules.get(constants.EGRESS_DIRECTION),
            port_id, network_id, fip_id=fip_id, ip_address=ip_address)
        ingress_ovn_rule = self.qos_driver._ovn_qos_rule(
            constants.INGRESS_DIRECTION,
            rules.get(constants.INGRESS_DIRECTION), port_id, network_id,
            fip_id=fip_id, ip_address=ip_address)

        with self.nb_api.transaction(check_error=True):
            ls = self.qos_driver.nb_idl.lookup(
                'Logical_Switch', ovn_utils.ovn_name(network_id))
            try:
                lsp = self.qos_driver.nb_idl.lsp_get(port_id).execute(
                    check_error=True)
            except idlutils.RowNotFound:
                # A LSP is created only in the tests that apply QoS rules to
                # an internal port. Any L3 QoS test (router gateway port or
                # floating IP), won't have a LSP associated and won't check
                # min-rate rules.
                pass

            self.assertEqual(len(rules), len(ls.qos_rules))
            for rule in ls.qos_rules:
                if expected_ext_ids:
                    self.assertDictEqual(expected_ext_ids, rule.external_ids)
                ref_rule = (egress_ovn_rule if rule.direction == 'from-lport'
                            else ingress_ovn_rule)
                action = {}
                if 'dscp' in ref_rule:
                    action = {'dscp': ref_rule['dscp']}
                bandwidth = {}
                if 'rate' in ref_rule:
                    bandwidth['rate'] = ref_rule['rate']
                    if ref_rule.get('burst'):
                        bandwidth['burst'] = ref_rule['burst']
                self.assertIn(port_id, rule.match)
                self.assertEqual(action, rule.action)
                self.assertEqual(bandwidth, rule.bandwidth)
            min_rate = rules.get(constants.EGRESS_DIRECTION, {}).get(
                qos_constants.RULE_TYPE_MINIMUM_BANDWIDTH)
            if min_rate is not None and check_min_rate:
                min_ovn = lsp.options.get(ovn_const.LSP_OPTIONS_QOS_MIN_RATE)
                self.assertEqual(str(min_rate['min_kbps']), min_ovn)


class TestOVNClientQosExtension(TestOVNClientQosExtensionBase):

    def setUp(self, maintenance_worker=False):
        super(TestOVNClientQosExtension, self).setUp(
            maintenance_worker=maintenance_worker)
        self._add_logical_switch()
        self.qos_driver = qos_extension.OVNClientQosExtension(
            nb_idl=self.nb_api)
        self.gw_port_id = 'gw_port_id'
        self._mock_get_router = mock.patch.object(l3_db.L3_NAT_dbonly_mixin,
                                                  '_get_router')
        self.mock_get_router = self._mock_get_router.start()
        self.mock_get_router.return_value = {'gw_port_id': self.gw_port_id}
        self._mock_qos_rules = mock.patch.object(self.qos_driver,
                                                 '_qos_rules')
        self.mock_qos_rules = self._mock_qos_rules.start()
        self.fip = {'router_id': 'router_id', 'qos_policy_id': 'qos_policy_id',
                    'floating_network_id': self.network_1,
                    'id': 'fip_id', 'floating_ip_address': '1.2.3.4'}

    def _add_logical_switch(self):
        self.network_1 = 'network_1'
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.ls_add(ovn_utils.ovn_name(self.network_1)))

    def _add_logical_switch_port(self, port_id):
        with self.nb_api.transaction(check_error=True) as txn:
            txn.add(self.nb_api.lsp_add(
                ovn_utils.ovn_name(self.network_1), port_id,
                options={'requested-chassis': 'compute1'}))

    def test__update_port_qos_rules(self):
        port = 'port1'
        self._add_logical_switch_port(port)

        def update_and_check(qos_rules):
            with self.nb_api.transaction(check_error=True) as txn:
                _qos_rules = copy.deepcopy(qos_rules)
                for direction in constants.VALID_DIRECTIONS:
                    _qos_rules[direction] = _qos_rules.get(direction, {})
                self.mock_qos_rules.return_value = _qos_rules
                self.qos_driver._update_port_qos_rules(
                    txn, port, self.network_1, 'qos1', None)
            self._check_rules(qos_rules, port, self.network_1)

        update_and_check(QOS_RULES_0)
        update_and_check(QOS_RULES_1)
        update_and_check(QOS_RULES_2)
        update_and_check(QOS_RULES_3)
        update_and_check({})

    def _update_fip_and_check(self, fip, qos_rules):
        with self.nb_api.transaction(check_error=True) as txn:
            _qos_rules = copy.deepcopy(qos_rules)
            for direction in constants.VALID_DIRECTIONS:
                _qos_rules[direction] = _qos_rules.get(direction, {})
            self.mock_qos_rules.return_value = _qos_rules
            self.qos_driver.update_floatingip(txn, fip)
        self._check_rules(qos_rules, self.gw_port_id, self.network_1,
                          fip_id='fip_id', ip_address='1.2.3.4')

    def test_create_floatingip(self):
        self._update_fip_and_check(self.fip, QOS_RULES_1)

    def test_update_floatingip(self):
        fip_updated = copy.deepcopy(self.fip)
        fip_updated['qos_policy_id'] = 'another_qos_policy'
        self._update_fip_and_check(self.fip, QOS_RULES_1)
        self._update_fip_and_check(fip_updated, QOS_RULES_2)
        self._update_fip_and_check(fip_updated, QOS_RULES_3)
        self._update_fip_and_check(fip_updated, {})

    def test_delete_floatingip(self):
        self._update_fip_and_check(self.fip, QOS_RULES_1)
        fip_dict = {'floating_network_id': self.fip['floating_network_id'],
                    'id': self.fip['id']}
        self._update_fip_and_check(fip_dict, {})


class TestOVNClientQosExtensionEndToEnd(TestOVNClientQosExtensionBase):

    def setUp(self, maintenance_worker=False):
        super(TestOVNClientQosExtensionEndToEnd, self).setUp(
            maintenance_worker=maintenance_worker)
        self.qos_driver = self.l3_plugin._ovn_client._qos_driver
        self._mock_qos_rules = mock.patch.object(self.qos_driver, '_qos_rules')
        self.mock_qos_rules = self._mock_qos_rules.start()

    def _create_router(self, name, gw_info=None, az_hints=None):
        router = {'router':
                  {'name': name,
                   'admin_state_up': True,
                   'tenant_id': self._tenant_id}}
        if az_hints:
            router['router']['availability_zone_hints'] = az_hints
        if gw_info:
            router['router']['external_gateway_info'] = gw_info
        return self.l3_plugin.create_router(self.context, router)

    def _create_ext_network(self, name, net_type, physnet, seg,
                            gateway, cidr):
        arg_list = (pnet.NETWORK_TYPE, external_net.EXTERNAL,)
        net_arg = {pnet.NETWORK_TYPE: net_type,
                   external_net.EXTERNAL: True}
        if seg:
            arg_list = arg_list + (pnet.SEGMENTATION_ID,)
            net_arg[pnet.SEGMENTATION_ID] = seg
        if physnet:
            arg_list = arg_list + (pnet.PHYSICAL_NETWORK,)
            net_arg[pnet.PHYSICAL_NETWORK] = physnet
        network = self._make_network(self.fmt, name, True,
                                     as_admin=True,
                                     arg_list=arg_list, **net_arg)
        if cidr:
            self._make_subnet(self.fmt, network, gateway, cidr,
                              ip_version=constants.IP_VERSION_4)
        return network

    def test_create_router_gateway_ip_qos(self):
        _qos_rules = copy.deepcopy(QOS_RULES_1)
        for direction in constants.VALID_DIRECTIONS:
            _qos_rules[direction] = _qos_rules.get(direction, {})
        self.mock_qos_rules.return_value = _qos_rules

        network = self._create_ext_network(
            utils.get_rand_name(), 'flat', 'physnet4',
            None, "110.0.0.1", "110.0.0.0/24")
        gw_info = {'network_id': network['network']['id']}
        router = self._create_router(utils.get_rand_name(), gw_info=gw_info)

        self._check_rules(
            _qos_rules, router['gw_port_id'],
            network['network']['id'],
            check_min_rate=False,
            expected_ext_ids={
                ovn_const.OVN_ROUTER_ID_EXT_ID_KEY: router['id']})
        self.l3_plugin.delete_router(self.context, router['id'])

    def test_delete_router_gateway_ip_qos_rules_removed(self):
        _qos_rules = copy.deepcopy(QOS_RULES_1)
        for direction in constants.VALID_DIRECTIONS:
            _qos_rules[direction] = _qos_rules.get(direction, {})
        self.mock_qos_rules.return_value = _qos_rules

        network = self._create_ext_network(
            utils.get_rand_name(), 'flat', 'physnet4',
            None, "120.0.0.1", "120.0.0.0/24")
        gw_info = {'network_id': network['network']['id']}
        router = self._create_router(utils.get_rand_name(), gw_info=gw_info)

        self._check_rules(
            _qos_rules, router['gw_port_id'],
            network['network']['id'],
            check_min_rate=False,
            expected_ext_ids={
                ovn_const.OVN_ROUTER_ID_EXT_ID_KEY: router['id']})
        ls = self.qos_driver.nb_idl.lookup(
            'Logical_Switch', ovn_utils.ovn_name(network['network']['id']))
        self.assertNotEqual(
            ls.qos_rules,
            [])

        self.l3_plugin.delete_router(self.context, router['id'])
        self.assertEqual(
            [],
            ls.qos_rules)

    def test_update_gateway_ip_qos(self):
        network = self._create_ext_network(
            utils.get_rand_name(), 'flat', 'physnet4',
            None, "130.0.0.1", "130.0.0.0/24")
        gw_info = {'network_id': network['network']['id']}
        router = self._create_router(utils.get_rand_name(), gw_info=gw_info)

        ls = self.qos_driver.nb_idl.lookup(
            'Logical_Switch', ovn_utils.ovn_name(network['network']['id']))
        self.assertEqual(
            [],
            ls.qos_rules)

        def update_and_check(qos_rules):
            _qos_rules = copy.deepcopy(qos_rules)
            for direction in constants.VALID_DIRECTIONS:
                _qos_rules[direction] = _qos_rules.get(direction, {})
            self.mock_qos_rules.return_value = _qos_rules
            self.l3_plugin.update_router(
                self.context, router['id'],
                {'router': {'admin_state_up': False}})
            self.l3_plugin.update_router(
                self.context, router['id'],
                {'router': {'admin_state_up': True}})
            self._check_rules(
                qos_rules, router['gw_port_id'], network['network']['id'],
                check_min_rate=False,
                expected_ext_ids={
                    ovn_const.OVN_ROUTER_ID_EXT_ID_KEY: router['id']})

        update_and_check(QOS_RULES_0)
        update_and_check(QOS_RULES_1)
        update_and_check(QOS_RULES_2)
        update_and_check(QOS_RULES_3)
        update_and_check({})

        self.l3_plugin.delete_router(self.context, router['id'])



# Copyright 2021 Huawei, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from neutron_lib import constants
from oslo_utils import uuidutils

from neutron.tests.fullstack import base
from neutron.tests.fullstack.resources import environment
from neutron.tests.fullstack.resources import machine
from neutron.tests.unit import testlib_api

load_tests = testlib_api.module_load_tests


class LocalIPTestCase(base.BaseFullStackTestCase):
    number_of_hosts = 2

    scenarios = [
        ('with_conntrack_rules', {'firewall_driver': 'noop'}),
        # with ovs firewall driver - test config will set static_nat=True
        ('static_nat', {'firewall_driver': 'openvswitch'}),
    ]

    def setUp(self):
        host_desc = [
            environment.HostDescription(
                l2_agent_type=constants.AGENT_TYPE_OVS,
                firewall_driver=self.firewall_driver,
                l3_agent=False,
                dhcp_agent=False) for _ in range(self.number_of_hosts)]
        env_desc = environment.EnvironmentDescription(
            mech_drivers='openvswitch',
            local_ip_ext=True)
        env = environment.Environment(env_desc, host_desc)
        super(LocalIPTestCase, self).setUp(env)
        self.project_id = uuidutils.generate_uuid()

        self.network = self.safe_client.create_network(
            self.project_id, 'network-local-ip-test')
        self.subnet_v4 = self.safe_client.create_subnet(
            self.project_id, self.network['id'],
            cidr='10.0.0.0/24',
            gateway_ip='10.0.0.1',
            name='subnet-v4-test')

    def _prepare_vms(self):
        port1 = self.safe_client.create_port(
            self.project_id, self.network['id'],
            self.environment.hosts[0].hostname,
            device_owner="compute:test_local_ip")

        port2 = self.safe_client.create_port(
            self.project_id, self.network['id'],
            self.environment.hosts[0].hostname,
            device_owner="compute:test_local_ip")

        port3 = self.safe_client.create_port(
            self.project_id, self.network['id'],
            self.environment.hosts[1].hostname,
            device_owner="compute:test_local_ip")

        vm1 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.project_id,
                self.safe_client,
                neutron_port=port1))

        vm2 = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[0],
                self.network['id'],
                self.project_id,
                self.safe_client,
                neutron_port=port2))

        vm_diff_host = self.useFixture(
            machine.FakeFullstackMachine(
                self.environment.hosts[1],
                self.network['id'],
                self.project_id,
                self.safe_client,
                neutron_port=port3))
        return machine.FakeFullstackMachinesList([vm1, vm2, vm_diff_host])

    def _create_local_ip(self):
        return self.safe_client.create_local_ip(
            self.project_id, network_id=self.network['id'])

    def _associate_local_ip(self, local_ip_id, port_id):
        return self.safe_client.create_local_ip_association(
            local_ip_id, port_id)

    def test_vm_is_accessible_by_local_ip(self):
        vms = self._prepare_vms()
        vms.block_until_all_boot()
        # first check basic connectivity between VMs
        vms.ping_all()

        local_ip = self._create_local_ip()
        self._associate_local_ip(
            local_ip['id'], vms[0].neutron_port['id'])
        # VM on same host should have access to this Local IP
        vms[1].block_until_ping(local_ip['local_ip_address'])

        # VM on different host shouldn't have access to this Local IP
        vms[2].assert_no_ping(local_ip['local_ip_address'])

        # NOTE(ykarel): response comes from local_ip assigned to vm0,
        # as a workaround sleep for arp cache expiry. It requires
        # fix in Local IP feature
        time.sleep(31)
        # check that VMs can still access each other with fixed IPs
        vms.ping_all()



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import uuid

from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import constants
from neutron_lib.plugins.ml2 import ovs_constants as ovs_const
from oslo_config import cfg
from oslo_log import log

from neutron._i18n import _
from neutron.agent import securitygroups_rpc
from neutron.conf.plugins.ml2.drivers.openvswitch import mech_ovs_conf
from neutron.plugins.ml2.drivers import mech_agent
from neutron.services.logapi.drivers.openvswitch import driver as log_driver
from neutron.services.qos.drivers.openvswitch import driver as ovs_qos_driver

LOG = log.getLogger(__name__)

IPTABLES_FW_DRIVER_FULL = ("neutron.agent.linux.iptables_firewall."
                           "OVSHybridIptablesFirewallDriver")

mech_ovs_conf.register_ovs_mech_driver_opts()


class OpenvswitchMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using openvswitch L2 agent.

    The OpenvswitchMechanismDriver integrates the ml2 plugin with the
    openvswitch L2 agent. Port binding with this driver requires the
    openvswitch agent to be running on the port's host, and that agent
    to have connectivity to at least one segment of the port's
    network.
    """

    resource_provider_uuid5_namespace = uuid.UUID(
        '87ee7d5c-73bb-11e8-9008-c4d987b2a692')

    def __init__(self):
        sg_enabled = securitygroups_rpc.is_firewall_enabled()
        vif_details = {portbindings.CAP_PORT_FILTER: sg_enabled,
                       portbindings.VIF_DETAILS_CONNECTIVITY:
                           self.connectivity}
        # NOTE(moshele): Bind DIRECT (SR-IOV) port allows
        # to offload the OVS flows using tc to the SR-IOV NIC.
        # We are using OVS mechanism driver because the openvswitch (>=2.8.0)
        # support hardware offload via tc and that allow us to manage the VF by
        # OpenFlow control plane using representor net-device.
        supported_vnic_types = [portbindings.VNIC_NORMAL,
                                portbindings.VNIC_DIRECT,
                                portbindings.VNIC_SMARTNIC,
                                portbindings.VNIC_VHOST_VDPA,
                                ]
        prohibit_list = cfg.CONF.OVS_DRIVER.vnic_type_prohibit_list
        super(OpenvswitchMechanismDriver, self).__init__(
            constants.AGENT_TYPE_OVS,
            portbindings.VIF_TYPE_OVS,
            vif_details,
            supported_vnic_types=supported_vnic_types,
            vnic_type_prohibit_list=prohibit_list)

        ovs_qos_driver.register()
        log_driver.register()

    @property
    def connectivity(self):
        return portbindings.CONNECTIVITY_L2

    def get_allowed_network_types(self, agent):
        return (agent['configurations'].get('tunnel_types', []) +
                [constants.TYPE_LOCAL, constants.TYPE_FLAT,
                 constants.TYPE_VLAN])

    def get_mappings(self, agent):
        return agent['configurations'].get('bridge_mappings', {})

    def get_standard_device_mappings(self, agent):
        """Return the agent's bridge mappings in a standard way.

        The common format for OVS and SRIOv mechanism drivers:
        {'physnet_name': ['device_or_bridge_1', 'device_or_bridge_2']}

        :param agent: The agent
        :returns A dict in the format: {'physnet_name': ['bridge_or_device']}
        :raises ValueError: if there is no bridge_mappings key in
                            agent['configurations']
        """
        if 'bridge_mappings' in agent['configurations']:
            return {k: [v] for k, v in
                    agent['configurations']['bridge_mappings'].items()}
        else:
            raise ValueError(_('Cannot standardize bridge mappings of agent '
                               'type: %s'), agent['agent_type'])

    def check_vlan_transparency(self, context):
        """Currently Openvswitch driver doesn't support vlan transparency."""
        return False

    def bind_port(self, context):
        vnic_type = context.current.get(portbindings.VNIC_TYPE,
                                        portbindings.VNIC_NORMAL)
        profile = context.current.get(portbindings.PROFILE)
        capabilities = []
        if profile:
            capabilities = profile.get('capabilities', [])
        # TODO(sean-k-mooney): in the case of the Mellanox connectx6 dx and lx
        # nics vhost-vdpa is only supported in switchdev mode but that is not
        # strictly required by other vendors so we should ideally add a config
        # value to control checking of switchdev support per host via the
        # agent['configurations']
        if (vnic_type == portbindings.VNIC_DIRECT and
                'switchdev' not in capabilities):
            LOG.debug("Refusing to bind due to unsupported vnic_type: %s with "
                      "no switchdev capability", portbindings.VNIC_DIRECT)
            return
        super(OpenvswitchMechanismDriver, self).bind_port(context)

    def get_supported_vif_type(self, agent):
        caps = agent['configurations'].get('ovs_capabilities', {})
        if (any(x in caps.get('iface_types', []) for x
                in [ovs_const.OVS_DPDK_VHOST_USER,
                    ovs_const.OVS_DPDK_VHOST_USER_CLIENT]) and
                agent['configurations'].get('datapath_type') ==
                ovs_const.OVS_DATAPATH_NETDEV):
            return portbindings.VIF_TYPE_VHOST_USER
        return self.vif_type

    def get_vif_type(self, context, agent, segment):
        if (context.current.get(portbindings.VNIC_TYPE) ==
                portbindings.VNIC_DIRECT):
            return portbindings.VIF_TYPE_OVS
        return self.get_supported_vif_type(agent)

    def get_vhost_mode(self, iface_types):
        # NOTE(sean-k-mooney): this function converts the ovs vhost user
        # driver mode into the qemu vhost user mode. If OVS is the server,
        # qemu is the client and vice-versa.
        if (ovs_const.OVS_DPDK_VHOST_USER_CLIENT in iface_types):
            return portbindings.VHOST_USER_MODE_SERVER
        return portbindings.VHOST_USER_MODE_CLIENT

    def get_vif_details(self, context, agent, segment):
        vif_details = self._pre_get_vif_details(agent, context)
        self._set_bridge_name(context.current, vif_details, agent)
        return vif_details

    @staticmethod
    def _set_bridge_name(port, vif_details, agent):
        # REVISIT(rawlin): add BridgeName as a nullable column to the Port
        # model and simply check here if it's set and insert it into the
        # vif_details.

        def set_bridge_name_inner(bridge_name):
            vif_details[portbindings.VIF_DETAILS_BRIDGE_NAME] = bridge_name

        bridge_name = agent['configurations'].get('integration_bridge')
        if bridge_name:
            vif_details[portbindings.VIF_DETAILS_BRIDGE_NAME] = bridge_name

        registry.publish(
            ovs_const.OVS_BRIDGE_NAME, events.BEFORE_READ,
            set_bridge_name_inner,
            payload=events.EventPayload(None, metadata={'port': port}))

    def _pre_get_vif_details(self, agent, context):
        a_config = agent['configurations']
        vif_type = self.get_vif_type(context, agent, segment=None)
        if vif_type != portbindings.VIF_TYPE_VHOST_USER:
            details = {
                **self.vif_details,
                portbindings.OVS_HYBRID_PLUG: a_config.get(
                    portbindings.OVS_HYBRID_PLUG)}
        else:
            sock_path = self.agent_vhu_sockpath(agent, context.current['id'])
            caps = a_config.get('ovs_capabilities', {})
            mode = self.get_vhost_mode(caps.get('iface_types', []))
            details = {portbindings.CAP_PORT_FILTER: False,
                       portbindings.OVS_HYBRID_PLUG: False,
                       portbindings.VHOST_USER_MODE: mode,
                       portbindings.VHOST_USER_OVS_PLUG: True,
                       portbindings.VHOST_USER_SOCKET: sock_path}
        details[portbindings.OVS_DATAPATH_TYPE] = a_config.get(
            'datapath_type', ovs_const.OVS_DATAPATH_SYSTEM)
        return details

    @staticmethod
    def agent_vhu_sockpath(agent, port_id):
        """Return the agent's vhost-user socket path for a given port"""
        sockdir = agent['configurations'].get('vhostuser_socket_dir',
                                              ovs_const.VHOST_USER_SOCKET_DIR)
        sock_name = (constants.VHOST_USER_DEVICE_PREFIX + port_id)[:14]
        return os.path.join(sockdir, sock_name)

    @staticmethod
    def provider_network_attribute_updates_supported():
        return [provider_net.SEGMENTATION_ID]



# Copyright (c) 2015 Red Hat Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import queue
import sys
import threading

import httplib2
from oslo_config import cfg
from oslo_log import log as logging

from neutron._i18n import _
from neutron.agent.l3 import ha
from neutron.agent.linux import daemon
from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as agent_utils
from neutron.common import config
from neutron.common import utils as common_utils
from neutron.conf.agent.l3 import keepalived
from neutron import privileged


LOG = logging.getLogger(__name__)
INITIAL_STATE_READ_TIMEOUT = 10


class KeepalivedUnixDomainConnection(agent_utils.UnixDomainHTTPConnection):
    def __init__(self, *args, **kwargs):
        # Old style super initialization is required!
        agent_utils.UnixDomainHTTPConnection.__init__(
            self, *args, **kwargs)
        self.socket_path = (
            ha.L3AgentKeepalivedStateChangeServer.
            get_keepalived_state_change_socket_path(cfg.CONF))


class MonitorDaemon(daemon.Daemon):
    def __init__(self, pidfile, router_id, user, group, namespace, conf_dir,
                 interface, cidr):
        self.router_id = router_id
        self.namespace = namespace
        self.conf_dir = conf_dir
        self.interface = interface
        self.cidr = cidr
        self.monitor = None
        self.event_stop = threading.Event()
        self.event_started = threading.Event()
        self.queue = queue.Queue()
        self._initial_state = None
        super(MonitorDaemon, self).__init__(pidfile, uuid=router_id,
                                            user=user, group=group)

    @property
    def initial_state(self):
        return self._initial_state

    @initial_state.setter
    def initial_state(self, state):
        if not self._initial_state:
            LOG.debug('Initial status of router %s is %s', self.router_id,
                      state)
            self._initial_state = state

    def run(self):
        self._thread_initial_state = threading.Thread(
            target=self.handle_initial_state)
        self._thread_ip_monitor = threading.Thread(
            target=ip_lib.ip_monitor,
            args=(self.namespace, self.queue, self.event_stop,
                  self.event_started))
        self._thread_read_queue = threading.Thread(
            target=self.read_queue,
            args=(self.queue, self.event_stop, self.event_started))
        self._thread_initial_state.start()
        self._thread_ip_monitor.start()
        self._thread_read_queue.start()

        # NOTE(ralonsoh): if the initial status is not read in a defined
        # timeout, "backup" state is set.
        self._thread_initial_state.join(timeout=INITIAL_STATE_READ_TIMEOUT)
        if not self.initial_state:
            LOG.warning('Timeout reading the initial status of router %s, '
                        'state is set to "backup".', self.router_id)
            self.write_state_change('backup')
            self.notify_agent('backup')

        self._thread_read_queue.join()

    def read_queue(self, _queue, event_stop, event_started):
        event_started.wait()
        while not event_stop.is_set():
            try:
                event = _queue.get(timeout=2)
            except queue.Empty:
                event = None
            if not event:
                continue

            if event['name'] == self.interface and event['cidr'] == self.cidr:
                if event['event'] == 'added':
                    new_state = 'primary'
                else:
                    new_state = 'backup'
                self.write_state_change(new_state)
                self.notify_agent(new_state)

    def handle_initial_state(self):
        try:
            state = 'backup'
            cidr = common_utils.ip_to_cidr(self.cidr)
            # NOTE(ralonsoh): "get_devices_with_ip" without passing an IP
            # address performs one single pyroute2 command. Because the number
            # of interfaces in the namespace is reduced, this is faster.
            for address in ip_lib.get_devices_with_ip(self.namespace):
                if (address['name'] == self.interface and
                        address['cidr'] == cidr):
                    state = 'primary'
                    break

            if not self.initial_state:
                self.write_state_change(state)
                self.notify_agent(state)
        except Exception:
            if not self.initial_state:
                LOG.exception('Failed to get initial status of router %s',
                              self.router_id)

    def write_state_change(self, state):
        self.initial_state = state
        with open(os.path.join(
                self.conf_dir, 'state'), 'w') as state_file:
            state_file.write(state)
        LOG.debug('Wrote router %s state %s', self.router_id, state)

    def notify_agent(self, state):
        resp, content = httplib2.Http().request(
            # Note that the message is sent via a Unix domain socket so that
            # the URL doesn't matter.
            'http://127.0.0.1/',
            headers={'X-Neutron-Router-Id': self.router_id,
                     'X-Neutron-State': state,
                     'Connection': 'close'},
            connection_type=KeepalivedUnixDomainConnection)

        if resp.status != 200:
            raise Exception(_('Unexpected response: %s') % resp)

        LOG.debug('Notified agent router %s, state %s', self.router_id, state)

    def handle_sigterm(self, signum, frame):
        self.event_stop.set()
        self._thread_read_queue.join(timeout=5)
        super(MonitorDaemon, self).handle_sigterm(signum, frame)


def configure(conf):
    config.register_common_config_options()
    config.init(sys.argv[1:])
    conf.set_override('log_dir', cfg.CONF.conf_dir)
    conf.set_override('use_syslog', True)
    config.setup_logging()
    privileged.default.set_client_mode(False)


def main():
    keepalived.register_cli_l3_agent_keepalived_opts()
    keepalived.register_l3_agent_keepalived_opts()
    configure(cfg.CONF)
    MonitorDaemon(cfg.CONF.pid_file,
                  cfg.CONF.router_id,
                  cfg.CONF.user,
                  cfg.CONF.group,
                  cfg.CONF.namespace,
                  cfg.CONF.conf_dir,
                  cfg.CONF.monitor_interface,
                  cfg.CONF.monitor_cidr).start()



# Copyright (c) 2021 Red Hat Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_policy import policy as base_policy
from oslo_utils import uuidutils

from neutron import policy
from neutron.tests.unit.conf.policies import test_base as base


class RouterAPITestCase(base.PolicyBaseTestCase):

    def setUp(self):
        super(RouterAPITestCase, self).setUp()
        self.target = {'project_id': self.project_id}
        self.alt_target = {'project_id': self.alt_project_id}


class SystemAdminTests(RouterAPITestCase):

    def setUp(self):
        super(SystemAdminTests, self).setUp()
        self.context = self.system_admin_ctx

    def test_create_router(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router', self.alt_target)

    def test_create_router_distributed(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:distributed', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:distributed', self.alt_target)

    def test_create_router_ha(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:ha', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:ha', self.alt_target)

    def test_create_router_external_gateway_info(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:external_gateway_info',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:external_gateway_info',
            self.alt_target)

    def test_create_router_external_gateway_info_network_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:network_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:network_id',
            self.alt_target)

    def test_create_router_external_gateway_info_enable_snat(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:enable_snat',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:enable_snat',
            self.alt_target)

    def test_create_router_external_gateway_info_external_fixed_ips(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context,
            'create_router:external_gateway_info:external_fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context,
            'create_router:external_gateway_info:external_fixed_ips',
            self.alt_target)

    def test_create_router_enable_default_route_bfd(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:enable_default_route_bfd',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:enable_default_route_bfd',
            self.alt_target)

    def test_create_router_enable_default_route_ecmp(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:enable_default_route_ecmp',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'create_router:enable_default_route_ecmp',
            self.alt_target)

    def test_get_router(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_router', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_router', self.alt_target)

    def test_get_router_distributed(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_router:distributed', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_router:distributed', self.alt_target)

    def test_get_router_ha(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_router:ha', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_router:ha', self.alt_target)

    def test_get_routers_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_routers_tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'get_routers_tags', self.alt_target)

    def test_update_router(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router', self.alt_target)

    def test_update_router_distributed(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:distributed', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:distributed', self.alt_target)

    def test_update_router_ha(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:ha', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:ha', self.alt_target)

    def test_update_router_external_gateway_info(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:external_gateway_info',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:external_gateway_info',
            self.alt_target)

    def test_update_router_external_gateway_info_network_id(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:network_id',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:network_id',
            self.alt_target)

    def test_update_router_external_gateway_info_enable_snat(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:enable_snat',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:enable_snat',
            self.alt_target)

    def test_update_router_external_gateway_info_external_fixed_ips(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context,
            'update_router:external_gateway_info:external_fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context,
            'update_router:external_gateway_info:external_fixed_ips',
            self.alt_target)

    def test_update_router_enable_default_route_bfd(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:enable_default_route_bfd',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:enable_default_route_bfd',
            self.alt_target)

    def test_update_router_enable_default_route_ecmp(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:enable_default_route_ecmp',
            self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_router:enable_default_route_ecmp',
            self.alt_target)

    def test_update_routers_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_routers_tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'update_routers_tags', self.alt_target)

    def test_delete_router(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_router', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_router', self.alt_target)

    def test_delete_routers_tags(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_routers_tags', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'delete_routers_tags', self.alt_target)

    def test_add_router_interface(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'add_router_interface', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'add_router_interface', self.alt_target)

    def test_remove_router_interface(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'remove_router_interface', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'remove_router_interface', self.alt_target)


class SystemMemberTests(SystemAdminTests):

    def setUp(self):
        super(SystemMemberTests, self).setUp()
        self.context = self.system_member_ctx


class SystemReaderTests(SystemMemberTests):

    def setUp(self):
        super(SystemReaderTests, self).setUp()
        self.context = self.system_reader_ctx


class AdminTests(RouterAPITestCase):

    def setUp(self):
        super(AdminTests, self).setUp()
        self.context = self.project_admin_ctx

    def test_create_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_router', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'create_router', self.alt_target))

    def test_create_router_distributed(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'create_router:distributed', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'create_router:distributed', self.alt_target))

    def test_create_router_ha(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_router:ha', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'create_router:ha', self.alt_target))

    def test_create_router_external_gateway_info(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info',
                           self.alt_target))

    def test_create_router_external_gateway_info_network_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info:network_id',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info:network_id',
                           self.alt_target))

    def test_create_router_external_gateway_info_enable_snat(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info:enable_snat',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info:enable_snat',
                           self.alt_target))

    def test_create_router_external_gateway_info_external_fixed_ips(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_router:external_gateway_info:external_fixed_ips',
                self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'create_router:external_gateway_info:external_fixed_ips',
                self.alt_target))

    def test_update_router_enable_default_route_bfd(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_router:enable_default_route_bfd',
                self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_router:enable_default_route_bfd',
                self.alt_target))

    def test_update_router_enable_default_route_ecmp(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_router:enable_default_route_ecmp',
                self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_router:enable_default_route_ecmp',
                self.alt_target))

    def test_get_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_router', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_router', self.alt_target))

    def test_get_router_distributed(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'get_router:distributed', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'get_router:distributed', self.alt_target))

    def test_get_router_ha(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_router:ha', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_router:ha', self.alt_target))

    def test_get_routers_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_routers_tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'get_routers_tags', self.alt_target))

    def test_update_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_router', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_router', self.alt_target))

    def test_update_router_distributed(self):
        self.assertTrue(
            policy.enforce(
                self.context, 'update_router:distributed', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'update_router:distributed', self.alt_target))

    def test_update_router_ha(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_router:ha', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_router:ha', self.alt_target))

    def test_update_router_external_gateway_info(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info',
                           self.alt_target))

    def test_update_router_external_gateway_info_network_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info:network_id',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info:network_id',
                           self.alt_target))

    def test_update_router_external_gateway_info_enable_snat(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info:enable_snat',
                           self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info:enable_snat',
                           self.alt_target))

    def test_update_router_external_gateway_info_external_fixed_ips(self):
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_router:external_gateway_info:external_fixed_ips',
                self.target))
        self.assertTrue(
            policy.enforce(
                self.context,
                'update_router:external_gateway_info:external_fixed_ips',
                self.alt_target))

    def test_update_routers_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_routers_tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'update_routers_tags',
                           self.alt_target))

    def test_delete_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_router', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_router', self.alt_target))

    def test_delete_routers_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_routers_tags', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'delete_routers_tags',
                           self.alt_target))

    def test_add_router_interface(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'add_router_interface', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'add_router_interface', self.alt_target))

    def test_remove_router_interface(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'remove_router_interface', self.target))
        self.assertTrue(
            policy.enforce(self.context,
                           'remove_router_interface', self.alt_target))


class ProjectMemberTests(AdminTests):

    def setUp(self):
        super(ProjectMemberTests, self).setUp()
        self.context = self.project_member_ctx

    def test_create_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'create_router', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router', self.alt_target)

    def test_create_router_distributed(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:distributed', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:distributed', self.alt_target)

    def test_create_router_ha(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:ha', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:ha', self.alt_target)

    def test_create_router_external_gateway_info(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info',
                           self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info',
            self.alt_target)

    def test_create_router_external_gateway_info_network_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'create_router:external_gateway_info:network_id',
                           self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:network_id',
            self.alt_target)

    def test_create_router_external_gateway_info_enable_snat(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:enable_snat',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:enable_snat',
            self.alt_target)

    def test_create_router_external_gateway_info_external_fixed_ips(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'create_router:external_gateway_info:external_fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'create_router:external_gateway_info:external_fixed_ips',
            self.alt_target)

    def test_update_router_enable_default_route_bfd(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:enable_default_route_bfd',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:enable_default_route_bfd',
            self.alt_target)

    def test_update_router_enable_default_route_ecmp(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:enable_default_route_ecmp',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:enable_default_route_ecmp',
            self.alt_target)

    def test_get_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_router', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router', self.alt_target)

    def test_get_router_distributed(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router:distributed', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router:distributed', self.alt_target)

    def test_get_router_ha(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router:ha', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router:ha', self.alt_target)

    def test_get_routers_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'get_routers_tags', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_routers_tags', self.alt_target)

    def test_update_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_router', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router', self.alt_target)

    def test_update_router_distributed(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:distributed', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:distributed', self.alt_target)

    def test_update_router_ha(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:ha', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:ha', self.alt_target)

    def test_update_router_external_gateway_info(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info',
                           self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info',
            self.alt_target)

    def test_update_router_external_gateway_info_network_id(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'update_router:external_gateway_info:network_id',
                           self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:network_id',
            self.alt_target)

    def test_update_router_external_gateway_info_enable_snat(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:enable_snat',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:enable_snat',
            self.alt_target)

    def test_update_router_external_gateway_info_external_fixed_ips(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:external_gateway_info:external_fixed_ips',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:external_gateway_info:external_fixed_ips',
            self.alt_target)

    def test_update_routers_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'update_routers_tags', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_routers_tags', self.alt_target)

    def test_delete_router(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_router', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_router', self.alt_target)

    def test_delete_routers_tags(self):
        self.assertTrue(
            policy.enforce(self.context, 'delete_routers_tags', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_routers_tags', self.alt_target)

    def test_add_router_interface(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'add_router_interface', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'add_router_interface', self.alt_target)

    def test_remove_router_interface(self):
        self.assertTrue(
            policy.enforce(self.context,
                           'remove_router_interface', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_router_interface', self.alt_target)


class ProjectReaderTests(ProjectMemberTests):

    def setUp(self):
        super(ProjectMemberTests, self).setUp()
        self.context = self.project_reader_ctx

    def test_create_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router', self.alt_target)

    def test_create_router_external_gateway_info(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info',
            self.alt_target)

    def test_create_router_external_gateway_info_network_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:network_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:network_id',
            self.alt_target)

    def test_update_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router', self.alt_target)

    def test_update_router_external_gateway_info(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info',
            self.alt_target)

    def test_update_router_external_gateway_info_network_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:network_id',
            self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:network_id',
            self.alt_target)

    def test_update_routers_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_routers_tags', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_routers_tags', self.alt_target)

    def test_delete_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_router', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_router', self.alt_target)

    def test_delete_routers_tags(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_routers_tags', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_routers_tags', self.alt_target)

    def test_add_router_interface(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'add_router_interface', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'add_router_interface', self.alt_target)

    def test_remove_router_interface(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_router_interface', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_router_interface', self.alt_target)


class ExtrarouteAPITestCase(base.PolicyBaseTestCase):

    def setUp(self):
        super(ExtrarouteAPITestCase, self).setUp()
        self.router = {
            'id': uuidutils.generate_uuid(),
            'project_id': self.project_id}

        self.target = {
            'project_id': self.project_id,
            'router_id': self.router['id'],
            'ext_parent_router_id': self.router['id']}
        self.alt_target = {
            'project_id': self.alt_project_id,
            'router_id': self.router['id'],
            'ext_parent_router_id': self.router['id']}


class SystemAdminExtrarouteTests(ExtrarouteAPITestCase):

    def setUp(self):
        super(SystemAdminExtrarouteTests, self).setUp()
        self.context = self.system_admin_ctx

    def test_add_extraroute(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'add_extraroutes', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'add_extraroutes', self.alt_target)

    def test_remove_extraroute(self):
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'remove_extraroutes', self.target)
        self.assertRaises(
            base_policy.InvalidScope,
            policy.enforce,
            self.context, 'remove_extraroutes', self.alt_target)


class SystemMemberExtrarouteTests(SystemAdminExtrarouteTests):

    def setUp(self):
        super(SystemMemberExtrarouteTests, self).setUp()
        self.context = self.system_member_ctx


class SystemReaderExtrarouteTests(SystemMemberExtrarouteTests):

    def setUp(self):
        super(SystemReaderExtrarouteTests, self).setUp()
        self.context = self.system_reader_ctx


class AdminExtrarouteTests(ExtrarouteAPITestCase):

    def setUp(self):
        super(AdminExtrarouteTests, self).setUp()
        self.context = self.project_admin_ctx

    def test_add_extraroute(self):
        self.assertTrue(
            policy.enforce(self.context, 'add_extraroutes', self.target))
        self.assertTrue(
            policy.enforce(self.context, 'add_extraroutes', self.alt_target))

    def test_remove_extraroute(self):
        self.assertTrue(
            policy.enforce(self.context, 'remove_extraroutes', self.target))
        self.assertTrue(
            policy.enforce(
                self.context, 'remove_extraroutes', self.alt_target))


class ProjectMemberExtrarouteTests(AdminExtrarouteTests):

    def setUp(self):
        super(ProjectMemberExtrarouteTests, self).setUp()
        self.context = self.project_member_ctx

    def test_add_extraroute(self):
        self.assertTrue(
            policy.enforce(self.context, 'add_extraroutes', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_extraroutes', self.alt_target)

    def test_remove_extraroute(self):
        self.assertTrue(
            policy.enforce(self.context, 'remove_extraroutes', self.target))
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_extraroutes', self.alt_target)


class ProjectReaderExtrarouteTests(ProjectMemberExtrarouteTests):

    def setUp(self):
        super(ProjectReaderExtrarouteTests, self).setUp()
        self.context = self.project_reader_ctx

    def test_add_extraroute(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'add_extraroutes', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'add_extraroutes', self.alt_target)

    def test_remove_extraroute(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_extraroutes', self.target)
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_extraroutes', self.alt_target)


class ServiceRoleTests(RouterAPITestCase):

    def setUp(self):
        super(ServiceRoleTests, self).setUp()
        self.context = self.service_ctx

    def test_create_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router', self.target)

    def test_create_router_distributed(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:distributed', self.target)

    def test_create_router_ha(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:ha', self.target)

    def test_create_router_external_gateway_info(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info',
            self.target)

    def test_create_router_external_gateway_info_network_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:network_id',
            self.target)

    def test_create_router_external_gateway_info_enable_snat(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'create_router:external_gateway_info:enable_snat',
            self.target)

    def test_create_router_external_gateway_info_external_fixed_ips(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'create_router:external_gateway_info:external_fixed_ips',
            self.target)

    def test_get_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router', self.target)

    def test_get_router_distributed(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router:distributed', self.target)

    def test_get_router_ha(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'get_router:ha', self.target)

    def test_update_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router', self.target)

    def test_update_router_distributed(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:distributed', self.target)

    def test_update_router_ha(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:ha', self.target)

    def test_update_router_external_gateway_info(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info',
            self.target)

    def test_update_router_external_gateway_info_network_id(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:network_id',
            self.target)

    def test_update_router_external_gateway_info_enable_snat(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'update_router:external_gateway_info:enable_snat',
            self.target)

    def test_update_router_external_gateway_info_external_fixed_ips(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context,
            'update_router:external_gateway_info:external_fixed_ips',
            self.target)

    def test_delete_router(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'delete_router', self.target)

    def test_add_router_interface(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'add_router_interface', self.target)

    def test_remove_router_interface(self):
        self.assertRaises(
            base_policy.PolicyNotAuthorized,
            policy.enforce,
            self.context, 'remove_router_interface', self.target)



# Copyright (c) 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
from unittest import mock

from neutron_lib import exceptions as n_exc
from neutron_lib.services.trunk import constants
from oslo_db import exception as obj_exc
from oslo_utils import uuidutils

from neutron.objects.db import api as obj_db_api
from neutron.objects import trunk as t_obj
from neutron.services.trunk import exceptions as t_exc
from neutron.tests.unit.objects import test_base
from neutron.tests.unit import testlib_api


class SubPortObjectTestCase(test_base.BaseObjectIfaceTestCase):

    _test_class = t_obj.SubPort

    def test_create_duplicates(self):
        with mock.patch.object(obj_db_api, 'create_object',
                               side_effect=obj_exc.DBDuplicateEntry):
            obj = self._test_class(self.context, **self.obj_fields[0])
            self.assertRaises(t_exc.DuplicateSubPort, obj.create)


class SubPortDbObjectTestCase(test_base.BaseDbObjectTestCase,
                              testlib_api.SqlTestCase):

    _test_class = t_obj.SubPort

    def setUp(self):
        super(SubPortDbObjectTestCase, self).setUp()
        self._network_id = self._create_test_network_id()
        for obj in self.obj_fields:
            self._create_test_port(
                id=obj['port_id'], network_id=self._network_id)
            self._create_trunk(trunk_id=obj['trunk_id'])

    def _create_trunk(self, trunk_id):
        port_id = self._create_test_port_id(network_id=self._network_id)
        trunk = t_obj.Trunk(self.context, id=trunk_id, port_id=port_id)
        trunk.create()

    def test_create_port_not_found(self):
        obj = self.obj_fields[0]
        obj['port_id'] = uuidutils.generate_uuid()

        sub_port = self._make_object(obj)
        self.assertRaises(n_exc.PortNotFound, sub_port.create)

    def test_create_trunk_not_found(self):
        obj = self.obj_fields[0]
        obj['trunk_id'] = uuidutils.generate_uuid()

        sub_port = self._make_object(obj)
        self.assertRaises(t_exc.TrunkNotFound, sub_port.create)


class TrunkObjectTestCase(test_base.BaseObjectIfaceTestCase):

    _test_class = t_obj.Trunk


class TrunkDbObjectTestCase(test_base.BaseDbObjectTestCase,
                            testlib_api.SqlTestCase):

    _test_class = t_obj.Trunk

    def setUp(self):
        super(TrunkDbObjectTestCase, self).setUp()

        self._network_id = self._create_test_network_id()
        sub_ports = []
        for obj in self.db_objs:
            sub_ports.extend(obj['sub_ports'])

        for obj in itertools.chain(self.obj_fields, sub_ports):
            self._create_test_port(
                id=obj['port_id'], network_id=self._network_id)

    def test_create_port_not_found(self):
        obj = self.obj_fields[0]
        obj['port_id'] = uuidutils.generate_uuid()

        trunk = self._make_object(obj)
        self.assertRaises(n_exc.PortNotFound, trunk.create)

    def _test_create_trunk_with_subports(self, port_id, vids):
        project_id = uuidutils.generate_uuid()

        sub_ports = []
        for vid in vids:
            vid_port_id = self._create_test_port_id(
                network_id=self._network_id)
            sub_ports.append(t_obj.SubPort(
                self.context, port_id=vid_port_id, segmentation_type='vlan',
                segmentation_id=vid))
        trunk = t_obj.Trunk(
            self.context, port_id=port_id, sub_ports=sub_ports,
            project_id=project_id)
        trunk.create()
        self.assertEqual(sub_ports, trunk.sub_ports)
        return trunk

    def test_create_with_sub_ports(self):
        trunk = self._test_create_trunk_with_subports(
            self.db_objs[0]['port_id'], [1, 2])

        def _as_tuple(sub_port):
            return (sub_port['port_id'],
                    sub_port['segmentation_type'],
                    sub_port['segmentation_id'])

        expected = {_as_tuple(port) for port in trunk.sub_ports}

        sub_ports = t_obj.SubPort.get_objects(self.context, trunk_id=trunk.id)
        self.assertEqual(expected, {_as_tuple(port) for port in sub_ports})

    def test_get_object_includes_correct_subports(self):
        trunk1_vids = [1, 2, 3]
        trunk2_vids = [4, 5, 6]
        port_id1 = self.db_objs[0]['port_id']
        trunk1 = self._test_create_trunk_with_subports(port_id1, trunk1_vids)

        port_id2 = uuidutils.generate_uuid()
        self._create_test_port(
            id=port_id2, network_id=self._network_id)
        self._test_create_trunk_with_subports(port_id2, trunk2_vids)

        listed_trunk1 = t_obj.Trunk.get_object(
            self.context,
            id=trunk1.id,
            port_id=port_id1
        )
        self.assertEqual(
            set(trunk1_vids),
            {sp.segmentation_id for sp in listed_trunk1.sub_ports}
        )

    def test_update_multiple_fields(self):
        trunk = t_obj.Trunk(context=self.context,
                            admin_state_up=False,
                            port_id=self.db_objs[0]['port_id'],
                            status=constants.TRUNK_DOWN_STATUS)
        trunk.create()
        fields = {'admin_state_up': True,
                  'status': constants.TRUNK_ACTIVE_STATUS}
        trunk.update(**fields)

        trunk = t_obj.Trunk.get_object(self.context, id=trunk.id)
        self._assert_trunk_attrs(trunk, **fields)

    def _assert_trunk_attrs(self, trunk, **kwargs):
        """Check the values passed in kwargs match the values of the trunk"""
        for k in trunk.fields:
            if k in kwargs:
                self.assertEqual(kwargs[k], trunk[k])

    def test_get_objects_tenant_id(self):
        trunk = t_obj.Trunk(context=self.context,
                            project_id='faketenant',
                            port_id=self.db_objs[0]['port_id'])
        trunk.create()
        self.assertIsNotNone(
            t_obj.Trunk.get_objects(self.context, tenant_id='faketenant'))

    def test_get_objects_both_tenant_and_project_ids(self):
        trunk = t_obj.Trunk(context=self.context,
                            project_id='faketenant',
                            port_id=self.db_objs[0]['port_id'])
        trunk.create()
        self.assertIsNotNone(
            t_obj.Trunk.get_objects(
                self.context, tenant_id='faketenant', project_id='faketenant'))



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions
from oslo_log import log as logging
from oslo_utils import versionutils

from neutron.agent.linux import utils as agent_utils

LOG = logging.getLogger(__name__)


# NOTE: Runtime checks are strongly discouraged in favor of sanity checks
#       which would be run at system setup time. Please consider writing a
#       sanity check instead.

def dnsmasq_host_tag_support():
    cmd = ['dnsmasq', '--test', '--dhcp-host=tag:foo']
    env = {'LC_ALL': 'C', 'PATH': '/sbin:/usr/sbin'}
    try:
        agent_utils.execute(cmd, addl_env=env, log_fail_as_error=False)
    except exceptions.ProcessExecutionError:
        return False
    return True


def get_keepalived_version():
    cmd = ['keepalived', '--version']
    env = {'LC_ALL': 'C', 'PATH': '/sbin:/usr/sbin'}
    try:
        # keepalived --version returns with stderr only
        res = agent_utils.execute(cmd, addl_env=env, log_fail_as_error=False,
                                  return_stderr=True)
        # First line is the interesting one here from stderr
        version_line = res[1].split('\n')[0]
        # Version string is of form 'v2.0.19', must remove 'v'
        keepalived_version = versionutils.convert_version_to_tuple(
            version_line.split()[1].lstrip('v'))
        return keepalived_version
    except exceptions.ProcessExecutionError:
        LOG.exception("Failed to get keepalived version")
        return False


def keepalived_use_no_track_support():

    keepalived_with_track = (2, 0, 3)
    keepalived_version = get_keepalived_version()
    if keepalived_version:
        return keepalived_version >= keepalived_with_track
    return False



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from neutron_lib import policy as neutron_policy
from oslo_log import versionutils
from oslo_policy import policy

from neutron.conf.policies import base


AG_COLLECTION_PATH = '/address-groups'
AG_RESOURCE_PATH = '/address-groups/{id}'

DEPRECATION_REASON = (
    "The Address scope API now supports system scope and default roles.")


rules = [
    policy.RuleDefault(
        'shared_address_groups',
        'field:address_groups:shared=True',
        'Definition of a shared address group'
    ),
    policy.DocumentedRuleDefault(
        name='get_address_group',
        check_str=neutron_policy.policy_or(
            base.ADMIN_OR_PROJECT_READER,
            'rule:shared_address_groups'),
        description='Get an address group',
        operations=[
            {
                'method': 'GET',
                'path': AG_COLLECTION_PATH,
            },
            {
                'method': 'GET',
                'path': AG_RESOURCE_PATH,
            },
        ],
        scope_types=['project'],
        deprecated_rule=policy.DeprecatedRule(
            name='get_address_group',
            check_str=neutron_policy.policy_or(
                neutron_policy.RULE_ADMIN_OR_OWNER,
                'rule:shared_address_groups'),
            deprecated_reason=DEPRECATION_REASON,
            deprecated_since=versionutils.deprecated.WALLABY)
    ),
]


def list_rules():
    return rules



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

import netaddr
from neutron_lib.api import validators
from neutron_lib import constants
from oslo_utils import netutils
from oslo_utils import uuidutils

from neutron._i18n import _
from neutron.common import utils as common_utils
from neutron.ipam import exceptions as ipam_exc
from neutron.ipam import utils as ipam_utils


class SubnetPool(object, metaclass=abc.ABCMeta):
    """Represents a pool of IPs available inside an address scope."""


class SubnetRequest(object, metaclass=abc.ABCMeta):
    """Carries the data needed to make a subnet request

    The data validated and carried by an instance of this class is the data
    that is common to any type of request.  This class shouldn't be
    instantiated on its own.  Rather, a subclass of this class should be used.
    """
    def __init__(self, tenant_id, subnet_id,
                 gateway_ip=None, allocation_pools=None):
        """Initialize and validate

        :param tenant_id: The tenant id who will own the subnet
        :type tenant_id: str uuid
        :param subnet_id: Neutron's subnet ID
        :type subnet_id: str uuid
        :param gateway_ip: An IP to reserve for the subnet gateway.
        :type gateway_ip: None or convertible to netaddr.IPAddress
        :param allocation_pools: The pool from which IPAM should allocate
            addresses.   The allocator *may* allow allocating addresses outside
            of this range if specifically requested.
        :type allocation_pools: A list of netaddr.IPRange.  None if not
            specified.
        """
        self._tenant_id = tenant_id
        self._subnet_id = subnet_id
        self._gateway_ip = None
        self._allocation_pools = None

        if gateway_ip is not None:
            self._gateway_ip = netaddr.IPAddress(gateway_ip)

        if allocation_pools is not None:
            allocation_pools = sorted(allocation_pools)
            previous = None
            for pool in allocation_pools:
                if not isinstance(pool, netaddr.ip.IPRange):
                    raise TypeError(_("Ranges must be netaddr.IPRange"))
                if previous and pool.first <= previous.last:
                    raise ValueError(_("Ranges must not overlap"))
                previous = pool
            if len(allocation_pools) > 1:
                # Checks that all the ranges are in the same IP version.
                # IPRange sorts first by ip version so we can get by with just
                # checking the first and the last range having sorted them
                # above.
                first_version = allocation_pools[0].version
                last_version = allocation_pools[-1].version
                if first_version != last_version:
                    raise ValueError(_("Ranges must be in the same IP "
                                       "version"))
            self._allocation_pools = allocation_pools

        if self.gateway_ip and self.allocation_pools:
            if self.gateway_ip.version != self.allocation_pools[0].version:
                raise ValueError(_("Gateway IP version inconsistent with "
                                   "allocation pool version"))

    @property
    def tenant_id(self):
        return self._tenant_id

    @property
    def subnet_id(self):
        return self._subnet_id

    @property
    def gateway_ip(self):
        return self._gateway_ip

    @property
    def allocation_pools(self):
        return self._allocation_pools

    def _validate_with_subnet(self, subnet_cidr):
        if self.allocation_pools:
            if subnet_cidr.version != self.allocation_pools[0].version:
                raise ipam_exc.IpamValueInvalid(_(
                    "allocation_pools use the wrong ip version"))
            for pool in self.allocation_pools:
                if pool not in subnet_cidr:
                    raise ipam_exc.IpamValueInvalid(_(
                        "allocation_pools are not in the subnet"))

    @staticmethod
    def _validate_gateway_ip_in_subnet(subnet_cidr, gateway_ip):
        """Validates if the Gateway IP is in subnet CIDR if needed

        If the Gateway (GW) IP address is in the subnet CIDR, we need to make
        sure that the user has not used the IPs reserved to represent the
        network or the broadcast domain.

        If the Gateway is not in the subnet CIDR, we do not validate it.
        Therefore, for such cases, it is assumed that its access is on link.
        """
        if not gateway_ip:
            return

        if ipam_utils.check_gateway_invalid_in_subnet(subnet_cidr, gateway_ip):
            raise ipam_exc.IpamValueInvalid(_(
                'Gateway IP %(gateway_ip)s cannot be the network or broadcast '
                'IP address %(subnet_cidr)s' % {'gateway_ip': gateway_ip,
                                                'subnet_cidr': subnet_cidr}))


class AnySubnetRequest(SubnetRequest):
    """A template for allocating an unspecified subnet from IPAM

    Support for this type of request in a driver is optional. For example, the
    initial reference implementation will not support this.  The API has no way
    of creating a subnet without a specific address until subnet-allocation is
    implemented.
    """
    WILDCARDS = {constants.IPv4: '0.0.0.0',
                 constants.IPv6: '::'}

    def __init__(self, tenant_id, subnet_id, version, prefixlen,
                 gateway_ip=None, allocation_pools=None):
        """Initialize AnySubnetRequest

        :param version: Either constants.IPv4 or constants.IPv6
        :param prefixlen: The prefix len requested.  Must be within the min and
            max allowed.
        :type prefixlen: int
        """
        super(AnySubnetRequest, self).__init__(
            tenant_id=tenant_id,
            subnet_id=subnet_id,
            gateway_ip=gateway_ip,
            allocation_pools=allocation_pools)

        net = netaddr.IPNetwork(self.WILDCARDS[version] + '/' + str(prefixlen))
        self._validate_with_subnet(net)

        self._prefixlen = prefixlen

    @property
    def prefixlen(self):
        return self._prefixlen


class SpecificSubnetRequest(SubnetRequest):
    """A template for allocating a specified subnet from IPAM

    The initial reference implementation will probably just allow any
    allocation, even overlapping ones.  This can be expanded on by future
    blueprints.
    """
    def __init__(self, tenant_id, subnet_id, subnet_cidr,
                 gateway_ip=None, allocation_pools=None):
        """Initialize SpecificSubnetRequest

        :param subnet: The subnet requested.  Can be IPv4 or IPv6.  However,
            when IPAM tries to fulfill this request, the IP version must match
            the version of the address scope being used.
        :type subnet: netaddr.IPNetwork or convertible to one
        """
        super(SpecificSubnetRequest, self).__init__(
            tenant_id=tenant_id,
            subnet_id=subnet_id,
            gateway_ip=gateway_ip,
            allocation_pools=allocation_pools)

        self._subnet_cidr = netaddr.IPNetwork(subnet_cidr)
        self._validate_with_subnet(self._subnet_cidr)
        self._validate_gateway_ip_in_subnet(self._subnet_cidr, self.gateway_ip)

    @property
    def subnet_cidr(self):
        return self._subnet_cidr

    @property
    def prefixlen(self):
        return self._subnet_cidr.prefixlen


class AddressRequest(object, metaclass=abc.ABCMeta):
    """Abstract base class for address requests"""


class SpecificAddressRequest(AddressRequest):
    """For requesting a specified address from IPAM"""
    def __init__(self, address):
        """Initialize SpecificAddressRequest

        :param address: The address being requested
        :type address: A netaddr.IPAddress or convertible to one.
        """
        super(SpecificAddressRequest, self).__init__()
        self._address = netaddr.IPAddress(address)

    @property
    def address(self):
        return self._address


class BulkAddressRequest(AddressRequest):
    """For requesting a batch of available addresses from IPAM"""
    def __init__(self, num_addresses):
        """Initialize BulkAddressRequest
        :param num_addresses: The quantity of IP addresses being requested
        :type num_addresses: int
        """
        super(BulkAddressRequest, self).__init__()
        self._num_addresses = num_addresses

    @property
    def num_addresses(self):
        return self._num_addresses


class AnyAddressRequest(AddressRequest):
    """Used to request any available address from the pool."""


class PreferNextAddressRequest(AnyAddressRequest):
    """Used to request next available IP address from the pool."""


class AutomaticAddressRequest(SpecificAddressRequest):
    """Used to create auto generated addresses, such as EUI64"""
    EUI64 = 'eui64'

    def _generate_eui64_address(self, **kwargs):
        if set(kwargs) != set(['prefix', 'mac']):
            raise ipam_exc.AddressCalculationFailure(
                address_type='eui-64',
                reason=_('must provide exactly 2 arguments - cidr and MAC'))
        prefix = kwargs['prefix']
        mac_address = kwargs['mac']
        return netutils.get_ipv6_addr_by_EUI64(prefix, mac_address)

    _address_generators = {EUI64: _generate_eui64_address}

    def __init__(self, address_type=EUI64, **kwargs):
        """Initialize AutomaticAddressRequest

        This constructor builds an automatic IP address. Parameter needed for
        generating it can be passed as optional keyword arguments.

        :param address_type: the type of address to generate.
            It could be an eui-64 address, a random IPv6 address, or
            an ipv4 link-local address.
            For the Kilo release only eui-64 addresses will be supported.
        """
        address_generator = self._address_generators.get(address_type)
        if not address_generator:
            raise ipam_exc.InvalidAddressType(address_type=address_type)
        address = address_generator(self, **kwargs)
        super(AutomaticAddressRequest, self).__init__(address)


class RouterGatewayAddressRequest(AddressRequest):
    """Used to request allocating the special router gateway address."""


class AddressRequestFactory(object):
    """Builds request using ip info

    Additional parameters(port and context) are not used in default
    implementation, but planned to be used in sub-classes
    provided by specific ipam driver,
    """

    @classmethod
    def get_request(cls, context, port, ip_dict):
        """Initialize AddressRequestFactory

        :param context: context (not used here, but can be used in sub-classes)
        :param port: port dict (not used here, but can be used in sub-classes)
        :param ip_dict: dict that can contain 'ip_address', 'mac' and
            'subnet_cidr' keys. Request to generate is selected depending on
             this ip_dict keys.
        :return: returns prepared AddressRequest (specific or any)
        """
        if ip_dict.get('ip_address'):
            return SpecificAddressRequest(ip_dict['ip_address'])
        elif ip_dict.get('eui64_address'):
            return AutomaticAddressRequest(prefix=ip_dict['subnet_cidr'],
                                           mac=ip_dict['mac'])
        elif (port['device_owner'] == constants.DEVICE_OWNER_DHCP or
              port['device_owner'] == constants.DEVICE_OWNER_DISTRIBUTED):
            # preserve previous behavior of DHCP ports choosing start of pool
            return PreferNextAddressRequest()
        else:
            return AnyAddressRequest()


class SubnetRequestFactory(object):
    """Builds request using subnet info"""

    @classmethod
    def get_request(cls, context, subnet, subnetpool):
        cidr = subnet.get('cidr')
        cidr = cidr if validators.is_attr_set(cidr) else None
        gateway_ip = subnet.get('gateway_ip')
        gateway_ip = gateway_ip if validators.is_attr_set(gateway_ip) else None
        subnet_id = subnet.get('id', uuidutils.generate_uuid())

        is_any_subnetpool_request = not (cidr or gateway_ip)
        if is_any_subnetpool_request:
            prefixlen = subnet['prefixlen']
            if not validators.is_attr_set(prefixlen):
                prefixlen = int(subnetpool['default_prefixlen'])

            return AnySubnetRequest(
                subnet['tenant_id'],
                subnet_id,
                common_utils.ip_version_from_int(subnetpool['ip_version']),
                prefixlen)
        else:
            alloc_pools = subnet.get('allocation_pools')
            alloc_pools = (alloc_pools if validators.is_attr_set(alloc_pools)
                           else None)
            if not cidr and gateway_ip:
                prefixlen = subnet['prefixlen']
                if not validators.is_attr_set(prefixlen):
                    prefixlen = int(subnetpool['default_prefixlen'])
                gw_ip_net = netaddr.IPNetwork('%s/%s' %
                                              (gateway_ip, prefixlen))
                cidr = gw_ip_net.cidr

            return SpecificSubnetRequest(
                subnet['tenant_id'],
                subnet_id,
                cidr,
                gateway_ip=gateway_ip,
                allocation_pools=alloc_pools)
